/***************************************************************************
 *   Copyright (C) 2004 by Kita Developers                                 *
 *   ikemo@users.sourceforge.jp                                            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include "bbsview.h"

#include <QtCore/QFile>
#include <QtCore/QRegExp>
#include <QtCore/QStringList>
#include <QtCore/QTextCodec>
#include <QtGui/QApplication>
#include <QtGui/QClipboard>
#include <QtGui/QHBoxLayout>
#include <QtGui/QHeaderView>
#include <QtGui/QMouseEvent>
#include <QtGui/QSpacerItem>
#include <QtGui/QTreeWidget>
#include <QtGui/QVBoxLayout>

#include <kaction.h>
#include <kcombobox.h>
#include <kconfig.h>
#include <klocale.h>
#include <kmenu.h>
#include <kmessagebox.h>
#include <krun.h>
#include <kstandarddirs.h>
#include <kio/netaccess.h>

#include "viewmediator.h"
#include "kitaui/listviewitem.h"
#include "libkita/boarddatabase.h"
#include "libkita/favoriteboards.h"
#include "libkita/globalconfig.h"
#include "libkita/kita_misc.h"

using namespace Kita;

struct Kita::Category
{
    QString category_name;
    QStringList boardNameList;
    QStringList boardUrlList;
};


/*--------------------------------------*/

BBSView::BBSView(QWidget *parent) : QWidget(parent), m_favorites(0)
{
    // copied from Base class
    bbsViewBaseLayout = new QVBoxLayout(this); 

    layout10 = new QHBoxLayout(0); 

    searchCombo = new KComboBox(this);
    searchCombo->setEditable(true);
    searchCombo->setMaxCount(10);
    layout10->addWidget(searchCombo);
    spacer2 = new QSpacerItem(467, 20, QSizePolicy::Expanding, QSizePolicy::Minimum);
    layout10->addItem(spacer2);
    bbsViewBaseLayout->addLayout(layout10);

    m_boardList = new QTreeWidget(this);
    m_boardList->setAlternatingRowColors(true);
    m_boardList->setRootIsDecorated(true);
    bbsViewBaseLayout->addWidget(m_boardList);
    resize(QSize(600, 482).expandedTo(minimumSizeHint()));
    // copy end

    m_boardList->setHeaderLabel(i18nc("@title:column", "Board Name"));
    m_boardList->header()->setClickable(false);

    connect(m_boardList, SIGNAL(itemActivated(QTreeWidgetItem*, int)),
            SLOT(loadBoard(QTreeWidgetItem*)));
    connect(FavoriteBoards::getInstance(), SIGNAL(changed()),
            SLOT(refreshFavoriteBoards()));
    connect(searchCombo, SIGNAL(textChanged(const QString&)),
             SLOT(filter(const QString&)));

    m_openWithBrowserAct
        = new KAction(i18nc("@action:inmenu", "Open with Web Browser"), this);
    m_copyUrlAct = new KAction(i18nc("@action:inmenu", "Copy URL"), this);
    m_copyTitleAndUrlAct
        = new KAction(i18nc("@action:inmenu", "Copy title and URL"), this);
    m_removeFromFavoritesAct
        = new KAction(i18nc("@action:inmenu", "Remove from Favorites"), this);
    m_addToFavoritesAct
        = new KAction(i18nc("@action:inmenu", "Add to Favorites"), this);
}

BBSView::~BBSView()
{
    saveOpened();
}

/**
 * download board list, parse, and write to "board_list"
 *
 * @see updateBoardList()
 * @see GlobalConfig::boardListURL()
 */
bool BBSView::downloadBoardList()
{
    // moved urls.
    QList<QString> oldUrls;
    QList<QString> newUrls;

    QString tmpFile;
    QString url = GlobalConfig::boardListUrl();
    if (! KIO::NetAccess::download(url, tmpFile, 0)) {
        return false;
    }

    QFile file(tmpFile);
    if (! file.open(QIODevice::ReadOnly)) {
        return false;
    }

    QTextStream stream(&file);
    QTextCodec* codec = QTextCodec::codecForName("Shift-JIS");
    stream.setCodec(codec);
    QString html = stream.readAll();

    // parse
    QStringList list;
    QList<Category> categoryList = getCategoryList(html);
    QList<Category>::iterator it;

    // Are there new boards or moved boards?
    QString newBoards;
    QString oldBoards;
    for (it = categoryList.begin(); it != categoryList.end(); ++it) {
        Category category = (*it);
        QList<QString> board_url_list = category.boardUrlList;
        if (board_url_list.isEmpty()) continue;

        int count = 0;
        for (QList<QString>::iterator it2 = board_url_list.begin();
                it2 != board_url_list.end(); ++it2) {

            QString boardUrl = *it2;
            QString boardName = category.boardNameList[ count ];
            QString oldUrl;
            BoardDatabase db(boardUrl);
            int ret = db.recordBoard(
                    boardName, oldUrl, Board_Unknown, true /* test only */);
            if (ret == Board_recordNew) {
                newBoards += boardName + "  ( " + category.category_name
                    + " )  " + boardUrl + '\n';
            }
            if (ret == Board_recordMoved) {
                oldBoards += boardName + "  ( " + category.category_name
                    + " )  " + oldUrl + "  ->  " + boardUrl + '\n';
            }
            count++;
            oldUrls += oldUrl;
            newUrls += boardUrl;
        }
    }

    const int maxNewBoard = 64;

    // show new board names
    if (!newBoards.isEmpty() && newBoards.count("\n") < maxNewBoard) {

        QStringList boardList = newBoards.split('\n');
        KMessageBox::informationList(this,
                                      i18n("New boards:\n\n"),
                                      boardList, "Kita");
    }

    // show moved board names
    if (!oldBoards.isEmpty()) {

        QStringList boardList = oldBoards.split('\n');
        KMessageBox::informationList(this,
                i18n("These boards were moved:\n\n")
                + i18n("\nPlease create the backup of those caches.\n"),
                boardList, "Kita");
    }

    if ((!newBoards.isEmpty() && newBoards.count("\n") < maxNewBoard)
            || !oldBoards.isEmpty()) {

        int ret = KMessageBox::warningYesNoCancel(this,
                i18n("Do you really want to update board list?"), "kita",
                KStandardGuiItem::yes(),
                KGuiItem(i18nc("@action:button", "Copy")),
                KStandardGuiItem::cancel(),
                QString(), KMessageBox::Dangerous);
        if (ret == KMessageBox::No) {
            QString str = i18n("New boards:\n\n") + newBoards + '\n'
                        + i18n("These boards were moved:\n\n") + oldBoards;
            QApplication::clipboard() ->setText(str , QClipboard::Clipboard);
            QApplication::clipboard() ->setText(str , QClipboard::Selection);
        } else if (ret == KMessageBox::Cancel)
            return false;
    } else if (newBoards.isEmpty() && oldBoards.isEmpty()) {

        KMessageBox::information(this, i18n("no new boards"), "Kita");
        return false;
    }

    // if moved URL exists. move files.
    for(int i=0; i<oldUrls.count(); i++) {
        QString oldUrl = oldUrls[ i ];
        QString newUrl = newUrls[ i ];

//        qDebug("move %s -> %s", oldURL.latin1(), newURL.latin1());
        BoardDatabase db;
        db.moveBoard(oldUrl, newUrl);
    }

    // save config
    QString configPath = KStandardDirs::locateLocal("appdata", "board_list");
    KConfig config(configPath);
    for (it = categoryList.begin(); it != categoryList.end(); ++it) {
        Category category = (*it);
        QList<QString> board_url_list = category.boardUrlList;

        if (board_url_list.isEmpty()) {
            continue;
        }

        list << category.category_name;
        KConfigGroup group = config.group(category.category_name);

        QList<QString>::iterator it2;
        int count = 0;
        for (it2 = board_url_list.begin(); it2 != board_url_list.end(); ++it2) {
            QString key = QString("item%1").arg(count);
            QString boardUrl = *it2;
            QString boardName = category.boardNameList[ count ];
            QStringList tmpList;
            tmpList << boardUrl;
            tmpList << boardName;
            group.writeEntry(key, tmpList);
            count++;
        }
    }
    KConfigGroup group2 = config.group("");
    group2.writeEntry("Categories", list);

    QDateTime now = QDateTime::currentDateTime();
    QString logPath = KStandardDirs::locateLocal("appdata", "board_log.txt");

    QFile logfile(logPath);
    if (logfile.open(QIODevice::WriteOnly | QIODevice::Append)) {
        QTextStream stream(&logfile);
        stream.setCodec("UTF-8");

        stream << "Date: " << now.toString("yyyy/MM/dd hh:mm:ss") << endl;
        stream << i18n("New boards:\n\n") + newBoards + '\n';
        stream << i18n("These boards were moved:\n\n") + oldBoards;
        stream << "----------------------------------------" << endl;
        logfile.close();
    }

    return true;
}

void BBSView::updateBoardList()
{
    if (downloadBoardList()) showBoardList();
}

/**
  * parse "board list" and show board list. don't download.
  *
  * @see updateBoardList()
  * @see downloadBoardList()
  */
void BBSView::showBoardList()
{
    // clear list
    m_boardList->clear();
    m_favorites = 0;

    QString configPath = KStandardDirs::locateLocal("appdata", "board_list");
    KConfig config(configPath);
    QStringList category_list = config.group("").readEntry("Categories", QStringList());

    ListViewItem* previous = 0;
    QStringList::iterator it;
    for (it = category_list.begin(); it != category_list.end(); ++it) {
        QString category = (*it);

        KConfigGroup group = config.group(category);
        ListViewItem* categoryItem = new ListViewItem(
                m_boardList, previous, QStringList(category));
        ListViewItem* previousBoard = 0;

        for (int count = 0; ; count++) {
            QString key = QString("item%1").arg(count);
            QStringList value = group.readEntry(key, QStringList());
            if (value.count() != 2) {
                break;
            }
            QString boardUrl = value[ 0 ];
            QString boardName = value[ 1 ];
            if (boardUrl.contains('/') != 4 || boardUrl.right(1) != "/") {
                // OK: http://pc8.2ch.net/linux/
                // NG: http://www.machi.to/
                // NG: http://find.2ch.net/enq/board.php
                continue;
            }
            QString oldUrl;
            BoardDatabase db(boardUrl);
            db.recordBoard(boardName, oldUrl);
            db.loadBBSHistory();
            previousBoard = new ListViewItem(categoryItem, previousBoard,
                    QStringList() << boardName << boardUrl);
        }
        previous = categoryItem;
    }

    QString boardUrl = "http://jbbs.livedoor.jp/computer/18420/" ;
    QString boardName = i18nc("@item:inlistbox", "Kita Board");
    QString oldUrl;
    new ListViewItem(m_boardList, 0, QStringList() << boardName << boardUrl);
    BoardDatabase db(boardUrl);
    db.recordBoard(boardName, oldUrl);

    loadExtBoard();
    refreshFavoriteBoards();
}


/* private */
void BBSView::loadExtBoard()
{
    QString configPath = KStandardDirs::locateLocal("appdata", "extbrd.conf");
    QFile file(configPath);
    if (file.open(QIODevice::ReadOnly)) {

        QTextStream stream(&file);
        stream.setCodec("UTF-8");

        QStringList list;
        QString str;

        ListViewItem* categoryItem
            = new ListViewItem(m_boardList, QStringList("extboard"));

        while (!(str = stream.readLine()).isEmpty()) {
            if (! str.isEmpty()) {
                QStringList list = str.split("<>");

                if (list.size() == 2 || list.size() == 3) {
                    QString board_title = list[ 0 ];
                    QString board_url = list[ 1 ];

                    new ListViewItem(categoryItem,
                                QStringList() << board_title << board_url);
                    QString oldUrl;
                    int type = Board_Unknown;
                    if (list.size() == 3) type = list[ 2 ].toInt();
                    BoardDatabase db(board_url);
                    db.recordBoard(board_title, oldUrl, type);
                }
            }
        }

        file.close();
    }
}

QList<Category> BBSView::getCategoryList(const QString& html) const
{
    QList<Category> result;

    QStringList lines = html.split('\n');
    QStringList::iterator it;

    Category current_category;
    current_category.category_name.clear();
    current_category.boardNameList.clear();
    current_category.boardUrlList.clear();
    for (it = lines.begin(); it != lines.end(); ++it) {
        QString category_name = getCategory(*it);
        if (!category_name.isEmpty()) {
            if (!current_category.category_name.isEmpty()) {
                result.append(current_category);
            }
            current_category.category_name = category_name;
            current_category.boardNameList.clear();
            current_category.boardUrlList.clear();
        } else {
            QRegExp regexp("<A HREF=([^ ]*).*>(.*)</A>", Qt::CaseInsensitive);
            if (regexp.indexIn((*it)) != -1) {
                QString url = regexp.cap(1);
                QString title = regexp.cap(2);
                if (isBoardUrl(url)
                        && !current_category.category_name.isEmpty()) {
                    current_category.boardNameList.append(title);
                    current_category.boardUrlList.append(url);
                }
            }
        }
    }
    return result;
}

void BBSView::refreshFavoriteBoards()
{
    if (!m_favorites) {
        m_favorites = new ListViewItem(m_boardList, 0,
                QStringList(i18nc("@item:inlistbox", "Favorites")));
    }

    // remove all items.
    do {
        QTreeWidgetItem* child = m_favorites->child(0);
        delete child;
    } while (m_favorites->childCount() != 0);

    // insert items.
    QList<KUrl> boards = FavoriteBoards::boards();
    QList<KUrl>::iterator it;

    for (it = boards.begin(); it != boards.end(); ++it) {
        QString boardUrl = (*it).prettyUrl();
        BoardDatabase db(boardUrl);
        QString boardName = db.boardName();

        new ListViewItem(m_favorites, QStringList() << boardName << boardUrl);
    }
}

void BBSView::loadBoard(QTreeWidgetItem* item)
{
    if (! item) return ;

    QString boardName = item->text(0);
    BoardDatabase db(item->text(1));
    QString boardUrl = db.boardUrl();

    if (boardUrl.isEmpty()) {
        return ;
    }

    ViewMediator::getInstance()->openBoard(boardUrl);
}

void BBSView::setFont(const QFont& font)
{
    m_boardList->setFont(font);
    searchCombo->setFont(font);
}

/* public slot :*/ /* virtual */
void BBSView::setFocus()
{
    m_boardList->setFocus();
}

void BBSView::slotMenuOpenWithBrowser()
{
    QTreeWidgetItem* item = m_boardList->currentItem();
    QString boardName = item->text(0);
    KUrl boardUrl = item->text(1);
    KRun::runUrl(boardUrl, "text/html", 0);
}

void BBSView::contextMenuEvent(QContextMenuEvent* e)
{
    QPoint point = e->globalPos();
    QTreeWidgetItem* item
        = m_boardList->itemAt(m_boardList->viewport()->mapFromGlobal(point));
    if (item == 0) {
        return;
    }

    KMenu popup(0);
    popup.addAction(m_openWithBrowserAct);
    popup.addAction(m_copyUrlAct);
    popup.addAction(m_copyTitleAndUrlAct);
    item->parent() == m_favorites ? popup.addAction(m_removeFromFavoritesAct)
        : popup.addAction(m_addToFavoritesAct);

    QString boardName = item->text(0);
    KUrl boardUrl = item->text(1);
    BoardDatabase db(boardUrl);
    KUrl boardUrl_upToDate = db.boardUrl();
    QClipboard* clipboard = QApplication::clipboard();

    QAction *action = popup.exec(point);
    if (action == m_openWithBrowserAct) {
        KRun::runUrl(boardUrl, "text/html", this);
    } else if (action == m_copyUrlAct) {
        clipboard->setText(boardUrl_upToDate.prettyUrl(),
                QClipboard::Clipboard);
        clipboard->setText(boardUrl_upToDate.prettyUrl(),
                QClipboard::Selection);
    } else if (action == m_copyTitleAndUrlAct) {
        clipboard->setText(boardName + '\n' + boardUrl_upToDate.prettyUrl(),
                QClipboard::Clipboard);
        clipboard->setText(boardName + '\n' + boardUrl_upToDate.prettyUrl(),
                QClipboard::Selection);
    } else if (action == m_addToFavoritesAct) {
        FavoriteBoards::append(boardUrl_upToDate);
    } else if (action == m_removeFromFavoritesAct) {
        FavoriteBoards::remove(boardUrl);
    }
}

void BBSView::loadOpened()
{
    QString configPath
        = KStandardDirs::locateLocal("appdata", "board_state.conf");
    KConfig config(configPath);

    QStringList openedList
        = config.group("").readEntry("Opened", QStringList());

    for (int i = 0, j = m_boardList->topLevelItemCount(); i < j; i++) {
        QTreeWidgetItem* item = m_boardList->topLevelItem(i);
        QString categoryName = item->text(0);
        if (openedList.indexOf(categoryName) != -1) {
            item->setExpanded(true);
        }
    }
}

void BBSView::saveOpened()
{
    QStringList openedList;
    for (int i = 0, j = m_boardList->topLevelItemCount(); i < j; i++) {
        QTreeWidgetItem* item = m_boardList->topLevelItem(i);
        QString categoryName = item->text(0);
        if (item->isExpanded()) {
            openedList << categoryName;
        }
    }
    QString configPath
        = KStandardDirs::locateLocal("appdata", "board_state.conf");
    KConfig config(configPath);

    config.group("").writeEntry("Opened", openedList);
}

/**
 * search and show matched item.
 *
 * @param str string to search.
 */
void BBSView::filter(const QString& str)
{

    for (int i = 0, j = m_boardList->topLevelItemCount(); i < j; i++) {
        bool matched = false;
        QTreeWidgetItem* categoryItem = m_boardList->topLevelItem(i);

        for (int k = 0, l = categoryItem->childCount(); k < l; k++) {
            QTreeWidgetItem* boardItem = categoryItem->child(k);
            QString boardName = boardItem->text(0);
            if (boardName.contains(str, Qt::CaseInsensitive)) {
                boardItem->setHidden(false);
                matched = true;
            } else {
                boardItem->setHidden(true);
            }
        }
        categoryItem->setHidden(!matched);
    }
}


void BBSView::mousePressEvent(QMouseEvent *e)
{
    QTreeWidgetItem* item = m_boardList->itemAt(e->pos());
    if (!item) return;

    QString boardName = item->text(0);
    BoardDatabase db(item->text(1));
    QString boardUrl = db.boardUrl();
    if (boardUrl.isEmpty()) {
        item->setExpanded(!item->isExpanded());
        return ;
    }

    switch (e->button()) {
    case Qt::MidButton:
        ViewMediator::getInstance()->openBoard(boardUrl);
        break;
    case Qt::LeftButton:
        ViewMediator::getInstance()->openBoard(boardUrl);
        break;
    default:
        return;
    }
}
