/***************************************************************************
*   Copyright (C) 2004 by Kita Developers                                 *
*   ikemo@users.sourceforge.jp                                            *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
***************************************************************************/

#include "boardtabwidget.h"

#include <QtGui/QApplication>
#include <QtGui/QClipboard>

#include <kaction.h>
#include <kactioncollection.h>
#include <kmenu.h>
#include <krun.h>
#include <kstandardshortcut.h>

#include "boardview.h"
#include "favoritelistview.h"
#include "viewmediator.h"
#include "libkita/boarddatabase.h"

using namespace Kita;

BoardTabWidget::BoardTabWidget(QWidget* parent) : TabWidgetBase(parent)
{
    setXMLFile("boardtabwidgetui.rc");

    setTabBar(new SubjectTabBar(this));

    FavoriteListView* favoriteList = new FavoriteListView(this);
    favoriteList->setObjectName("favoriteTab");
    ViewMediator::getInstance()->setFavoriteListView(favoriteList);
    addTab(favoriteList, i18nc("@title:tab Favorite boards", "Favorite"));

    setupActions();
}

BoardTabWidget::~BoardTabWidget()
{}

void BoardTabWidget::updateBoardView(const KUrl& datUrl)
{
    for(int i=0; i<count(); i++) {
        BoardView* view = isSubjectView(widget(i));
        if(view) {
            view->slotUpdateSubject(datUrl);
        }
    }
}

/* public slots */
void BoardTabWidget::loadBoard(const KUrl& boardUrl)
{
    BoardView * view = findView(boardUrl);
    BoardDatabase db(boardUrl);
    QString boardName = db.boardName();
    if (!view) {
        view = createView(boardName);
    }

    if (view) {
        setCurrentWidget(view);
        view->loadBoard(boardUrl);
    }
}

/* create BoardView */ /* private */
BoardView* BoardTabWidget::createView(QString label)
{
    BoardView * view = new BoardView(this);
    if (view) {
        insertTab(count() - 1, view, label);
    }

    return view;
}

/* private */
BoardView* BoardTabWidget::findView(const KUrl& boardUrl)
{
    int max = count() - 1;
    if (max <= 0) return 0;
    int i = 0;

    while (i < max) {
        BoardView * view = isSubjectView(widget(i));
        if (view && view->boardUrl() == boardUrl) return view;
        i++;
    }

    return 0;
}


/* protected */ /* virtual */
void BoardTabWidget::deleteWidget(QWidget* w)
{
    BoardView * view = isSubjectView(w);
    if (!view) return ;
    removePage(view);
    delete view;
}


/* private */
BoardView* BoardTabWidget::isSubjectView(QWidget* w)
{
    return qobject_cast<BoardView*>(w);
}

/*--------------------------------*/
/* BoardView actions */

/* private */
void BoardTabWidget::setupActions()
{
    KAction* find_action = actionCollection()->addAction("subjectview_find");
    find_action->setText(i18n("Find"));
    find_action->setShortcut(KStandardShortcut::find());
    connect(find_action, SIGNAL(triggered()), this, SLOT(slotFocusSearchCombo()));

    KAction* reload_action = actionCollection()->addAction("subjectview_reload");
    reload_action->setText(i18nc("@action:button Reload the subject view",
            "Reload"));
    reload_action->setShortcut(KStandardShortcut::reload());
    connect(reload_action, SIGNAL(triggered()), this, SLOT(slotReloadButton()));

    KAction* showoldlogs_action = actionCollection()->addAction("subjectview_showoldlogs");
    showoldlogs_action->setText(i18n("Show Old Logs"));
    connect(showoldlogs_action, SIGNAL(triggered()), this, SLOT(slotShowOldLogs()));
}

/* public slot */
void BoardTabWidget::slotReloadButton()
{
    BoardView * view = isSubjectView(currentWidget());
    if (view) view->reloadSubject();
}

/* public slot */
void BoardTabWidget::slotFocusSearchCombo()
{
    BoardView * view = isSubjectView(currentWidget());
    if (view) view->slotFocusSearchCombo();
}

/* public slot */
void BoardTabWidget::slotShowOldLogs(int idx)
{
    BoardView * view;
    if (idx == -1) view = isSubjectView(currentWidget());
    else view = isSubjectView(widget(idx));
    if (view) view->toggleShowOldLogs();
}

/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/



SubjectTabBar::SubjectTabBar(QWidget* parent)
        : KTabBar(parent)
{
    connect(this, SIGNAL(contextMenu(int, const QPoint&)),
             SLOT(showPopupMenu(int, const QPoint&)));
    m_closeAct = new KAction(i18nc("@action:inmenu", "Close This Tab"), this);
    m_closeOtherAct
        = new KAction(i18nc("@action:inmenu", "Close Other Tabs"), this);
    m_closeRightAct
        = new KAction(i18nc("@action:inmenu", "Close Right Tabs"), this);
    m_closeLeftAct
        = new KAction(i18nc("@action:inmenu", "Close Left Tabs"), this);
    m_showOldLogsAct
        = new KAction(i18nc("@action:inmenu", "Show Old Logs"), this);
    m_openBrowserAct
        = new KAction(i18nc("@action:inmenu", "Open with Web Browser"), this);
    m_copyTitleAct
        = new KAction(i18nc("@action:inmenu", "Copy Title and URL"), this);
}

SubjectTabBar::~SubjectTabBar()
{}


/* private */ /* virtual */
void SubjectTabBar::showPopupMenu(int idx, const QPoint& global)
{
    BoardTabWidget* tabwidget = static_cast<BoardTabWidget*>(parentWidget());
    KActionCollection * collection = tabwidget->actionCollection();

    if (QString::compare(tabwidget->widget(idx)->objectName(), "favoriteTab")
            == 0)
        return;
    KMenu popup(this);
    popup.addAction(m_closeAct);
    popup.addAction(collection->action("tab_prevtab"));
    popup.addAction(collection->action("tab_nexttab"));
    popup.addSeparator();

    popup.addAction(m_closeOtherAct);
    popup.addAction(m_closeRightAct);
    popup.addAction(m_closeLeftAct);
    popup.addAction(collection->action("tab_closealltab"));
    popup.addSeparator();

    popup.addAction(m_showOldLogsAct);
    popup.addAction(m_openBrowserAct);
    popup.addAction(m_copyTitleAct);
    popup.addSeparator();

    popup.addAction(collection->action("tab_configkeys"));

    BoardView* subjectView = static_cast<BoardView *>(tabwidget->widget(idx));
    QClipboard* clipboard = QApplication::clipboard();

    QAction* action = popup.exec(global);
    if (action == m_closeAct) {
        tabwidget->slotCloseTab(idx);
    } else if (action == m_closeOtherAct) {
        tabwidget->slotCloseOtherTab(idx);
    } else if (action == m_closeRightAct) {
        tabwidget->slotCloseRightTab(idx);
    } else if (action == m_closeLeftAct) {
        tabwidget->slotCloseLeftTab(idx);
    } else if (action == m_showOldLogsAct) {
        tabwidget->slotShowOldLogs(idx);
    } else if (action == m_openBrowserAct) {
        KRun::runUrl(subjectView->boardUrl(), "text/html", this);
    } else if (action == m_copyTitleAct) {
        BoardDatabase db(subjectView->boardUrl());
        QString cliptxt = db.boardName()
            + '\n' + subjectView->boardUrl().prettyUrl();
        clipboard->setText(cliptxt , QClipboard::Clipboard);
        clipboard->setText(cliptxt , QClipboard::Selection);
    }
}
