/***************************************************************************
*   Copyright (C) 2004 by Kita Developers                                 *
*   ikemo@users.sourceforge.jp                                            *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
***************************************************************************/

#include "boardview.h"

#include <QtCore/QDateTime>
#include <QtCore/QList>
#include <QtGui/QApplication>
#include <QtGui/QMouseEvent>

#include <kaction.h>
#include <kconfig.h>
#include <kmessagebox.h>
#include <kstandarddirs.h>

#include "boardtabwidget.h"
#include "threadlistheaderview.h"
#include "threadlistviewitem.h"
#include "viewmediator.h"
#include "libkita/boarddatabase.h"
#include "libkita/datmanager.h"
#include "libkita/globalconfig.h"
#include "libkita/kita_misc.h"
#include "libkita/thread.h"

using namespace Kita;

BoardView::BoardView(BoardTabWidget* parent) : ThreadListView(parent)
{
    init();
    m_parent = parent;
    closeButton->setEnabled(true);

    connect(reloadButton, SIGNAL(clicked()),
             SLOT(reloadSubject()));

    connect(closeButton, SIGNAL(clicked()),
             SLOT(slotCloseButton()));

    //connect(subjectList, SIGNAL(columnResized(int, int, int)),
    //         SLOT(slotSizeChange(int, int, int))); TODO

    loadLayout();
    loadHeaderOnOff();

    m_deleteLogAct->setVisible(true);
}

/* public */
void BoardView::init()
{
    m_unreadNum = 0;
    m_readNum = 0;
    m_newNum = 0;
    m_showOldLogs = false;
    m_enableSizeChange = false;
}


void BoardView::reloadSubject()
{
    if (!m_boardUrl.isEmpty()) {
        loadBoard(m_boardUrl);
    }
}


/* private slot */
void BoardView::slotCloseButton()
{
    m_parent->slotCloseCurrentTab();
}


const KUrl BoardView::boardUrl()
{
    return m_boardUrl;
}

void BoardView::toggleShowOldLogs()
{
    m_showOldLogs = !m_showOldLogs;
    loadBoard(m_boardUrl, false);
}

// used to sort rows
enum {
    ThreadOld,
    ThreadReaded,
    ThreadNormal,
    ThreadNew,
    ThreadRead,
    ThreadHasUnread
};

void BoardView::loadBoard(const KUrl& url, bool online)
{
    activateWindow();
    topLevelWidget()->raise();
    m_enableSizeChange = false;

    // reset member variables.
    // FIXME: FavoriteListView::update()
    {
        m_hitList.clear();
        m_nextHitIndex = 0;
        m_prevquery = QStringList();
    }

    m_boardUrl = url;
    m_unreadNum = 0;
    m_readNum = 0;
    m_newNum = 0;

    // get list of pointers of Thread classes
    QList<Thread*> oldLogList;
    QList<Thread*> threadList;
    BoardDatabase db(m_boardUrl);
    db.getThreadList(m_showOldLogs, online, threadList, oldLogList);

    // reset list
    subjectList->clearContents();

    QDateTime current = QDateTime::currentDateTime();
    int countNew = threadList.count();
    int countOld = oldLogList.count();
    int count = countNew + countOld;
    subjectList->setRowCount(count);
    subjectList->setSortingEnabled(false);
    for (int i = 0; i < count; i++) {
        Thread* thread
            = i < countNew ? threadList.at(i) : oldLogList.at(i - countNew);
        KUrl datUrl = thread->datUrl();
        for (int j = 0, k = subjectList->columnCount(); j < k; j++) {
            ThreadListViewItem* item = new ThreadListViewItem(j);
            subjectList->setItem(i, j, item);
        }
        int id = (i < countNew ? i + 1 : 0);
        int order = i + 1;
        updateRow(i, datUrl, current, id, order);
    }
    subjectList->setSortingEnabled(true);

    if (hideButton->isChecked()) {
        hideButton->toggle();
    }
    ViewMediator::getInstance()->setMainUrlLine(m_boardUrl);

    switch (GlobalConfig::listSortOrder()) {
    case GlobalConfig::EnumListSortOrder::Mark:
        subjectList->sortItems(ColumnMark);
        break;
    case GlobalConfig::EnumListSortOrder::ID:
        subjectList->sortItems(ColumnId);
        break;
    default:
        // do nothing
        break;
    }

    subjectList->setFocus();
    updateKindLabel();

    // restore column size
    loadLayout();
    loadHeaderOnOff();
    m_enableSizeChange = true;
    for (int i = 0, j = subjectList->columnCount(); i < j; i++) {
        subjectList->resizeColumnToContents(i);
    }
    for (int i = 0, j = subjectList->rowCount(); i < j; i++) {
        subjectList->resizeRowToContents(i);
    }
}

/* public slot */ /* virtual */
void BoardView::setFocus()
{
    subjectList->setFocus();
}

/* public slot */
void BoardView::slotFocusSearchCombo()
{
    searchCombo->hasFocus() ? setFocus() : searchCombo->setFocus();
}

void BoardView::updateKindLabel()
{
    QString fmtstr;
    fmtstr += QString("<font color=#C50000>%1</font>").arg(m_unreadNum);
    fmtstr += QString("/<font color=#00C200>%1</font>").arg(m_readNum);
    fmtstr += QString("/<font color=#5AAAFF>%1</font>").arg(m_newNum);
    kindLabel->setText(fmtstr);
}

void BoardView::setFont(const QFont& font)
{
    subjectList->setFont(font);
    searchCombo->setFont(font);
}

void BoardView::slotUpdateSubject(const KUrl& url)
{
    QDateTime current = QDateTime::currentDateTime();

    KUrl datUrl = getDatUrl(url);
    for (int i = 0, j = subjectList->rowCount(); i < j; i++) {
        if (subjectList->item(i, ColumnDatUrl)->text() == datUrl.prettyUrl()) {

            switch (subjectList->item(i, ColumnMark)->data(Qt::UserRole)
                    .toInt()) {
            case ThreadReaded:
            case ThreadRead:
                m_readNum--;
                break;
            case ThreadNew:
                m_newNum--;
                break;
            case ThreadHasUnread:
                m_unreadNum--;
                break;
            }

            Thread* thread = Thread::getByUrlNew(datUrl);
            if (thread == 0)
                return;
            int id = subjectList->item(i, ColumnId)->text().toInt();
            int order = subjectList->item(i, ColumnIdOrder)->text().toInt();
            updateRow(i, datUrl, current, id, order);

            updateKindLabel();

            return;
        }
    }
}

/* if id == 0, this thread is old thread. */
/* private */
void BoardView::updateRow(int row, const KUrl& datUrl,
        const QDateTime& current, int id, int order)
{
    QDateTime since;
    since.setTime_t(datToSince(datUrl));
    DatManager datManager(datUrl);
    QString threadName = datManager.threadName();
    int resNum = datManager.getResNum();
    int readNum = datManager.getReadNum();
    int viewPos = datManager.getViewPos();
    double speed = resNum / (since.secsTo(current) / (60.0 * 60.0 * 24.0));

    if (id)
        subjectList->item(row, ColumnId)->setText(QString::number(id));
    subjectList->item(row, ColumnIdOrder)->setText(QString::number(order));
    subjectList->item(row, ColumnSubject)->setText(threadName);
    subjectList->item(row, ColumnResNum)->setText(QString("%1").arg(resNum, 4));
    subjectList->item(row, ColumnReadNum)
        ->setText((viewPos > 0) ? QString("%1").arg(viewPos, 4) : QString());
    subjectList->item(row, ColumnUnread)
        ->setText((viewPos > 0 && resNum > viewPos)
                ? QString("%1").arg(resNum - viewPos, 4) : QString());
    subjectList->item(row, ColumnSince)
        ->setText(since.toString("yy/MM/dd hh:mm"));
    subjectList->item(row, ColumnDatUrl)->setText(datUrl.prettyUrl());
    subjectList->item(row, ColumnSpeed)
        ->setText(QString(" %1 ").arg(speed, 0, 'f', 2));

    // set mark order
    QTableWidgetItem *item = subjectList->item(row, ColumnMark);
    int orderValue;
    if (!id) { // old thread
        orderValue = ThreadOld;
    } else if (readNum > 0 && (resNum > readNum || resNum > viewPos)) {
        // There are new responses.
        item->setIcon(QIcon(SmallIcon("unread")));
        m_unreadNum++;
        orderValue = ThreadHasUnread;
    } else if (readNum > 0) { // Cache exists
        item->setIcon(QIcon(SmallIcon("read")));
        m_readNum++;
        orderValue = (viewPos > 1000) ? ThreadReaded : ThreadRead;
    } else if (since.secsTo(current) < 3600 * GlobalConfig::markTime()) {
        // new thread
        item->setIcon(QIcon(SmallIcon("newthread")));
        m_newNum++;
        orderValue = ThreadNew;
    } else { // normal
        item->setIcon(QIcon());
        orderValue = ThreadNormal;
    }
    item->setData(Qt::UserRole, orderValue);

    // no effect: m_unreadNum, m_readNum, m_newNum, markOrder
    if (datManager.isMainThreadOpened() && resNum == readNum) {
        // opened
        subjectList->item(row, ColumnMark)->setIcon(QIcon(SmallIcon("open")));
    }
}

void BoardView::deleteLog(const KUrl& url)
{
    if (KMessageBox::warningYesNo(this, i18n("Do you want to delete Log?"),
            "Kita") == KMessageBox::Yes) {
        if (DatManager(url).deleteCache()) {
            ViewMediator::getInstance()->closeThreadTab(url);
            slotUpdateSubject(url);
        }
    }
}

void BoardView::slotSizeChange(int section, int oldSize, int newSize)
{
    if (!m_enableSizeChange) return ;
    if (autoResize()) return ;

    QString configPath
        = KStandardDirs::locateLocal("appdata", "subjectview.conf");
//    KConfig config(configPath); TODO
//    subjectList->saveLayout(&config, "Layout"); TODO
}

void BoardView::loadLayout()
{
    if (autoResize()) return ;
    QString configPath
        = KStandardDirs::locateLocal("appdata", "subjectview.conf");
    KConfig config(configPath);
    // subjectList->restoreLayout(&config, "Layout"); TODO
}

void BoardView::saveHeaderOnOff()
{
    QString configPath
        = KStandardDirs::locateLocal("appdata", "subjectview.conf");
    KConfig config(configPath);

    KConfigGroup group = config.group("Column");
    for (int i = ColumnBegin; i <= ColumnEnd; i++) {
        group.writeEntry(ThreadListHeaderView::attribute[i].keyName,
                subjectList->columnWidth(i) != 0);
    }
}

void BoardView::loadHeaderOnOff()
{
    QString configPath
        = KStandardDirs::locateLocal("appdata", "subjectview.conf");
    KConfig config(configPath);

    KConfigGroup group = config.group("Column");
    for (int i = ColumnBegin; i <= ColumnEnd; i++) {
        bool isShown = group.readEntry(
                ThreadListHeaderView::attribute[i].keyName,
                ThreadListHeaderView::attribute[i].showDefault);
        isShown ? showColumn(i) : hideColumn(i);
    }
}

bool BoardView::autoResize()
{
    QString configPath
        = KStandardDirs::locateLocal("appdata", "subjectview.conf");
    KConfig config(configPath);

    return config.group("Column").readEntry("AutoResize", true);
}

void BoardView::setAutoResize(bool flag)
{
    QString configPath
        = KStandardDirs::locateLocal("appdata", "subjectview.conf");
    KConfig config(configPath);

    config.group("Column").writeEntry("AutoResize", flag);
}
