/***************************************************************************
 *   Copyright (C) 2004, 2007 by Kita Developers                           *
 *   ikemo@users.sourceforge.jp                                            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

/* This class manages the DOM elements  */

#include "domtree.h"

#include <kurl.h>
#include <dom/dom_text.h>

#include "libkita/datinfo.h"
#include "libkita/datmanager.h"
#include "libkita/kita_misc.h"
#include "libkita/kita-utf8.h"

using namespace Kita;

DomTree::DomTree(const DOM::HTMLDocument& hdoc, const KUrl& datUrl)
{
    m_hdoc = hdoc;
    m_bufSize = 0;
    m_bottomNum = 0;

    // create the nodes of footer, header, etc.
    createKokoyon();
    createFooter();
    createHeader();

    // get pointer of DatInfo
    // Note that m_datURL is already locked in the HTMLPart.
    m_datInfo = DatManager(datUrl).getDatInfoPointer();
}

DomTree::~DomTree() {}

/*
 * This function creates DOM elements of both title and body.
 * To show the res, call appendRes()
 */
bool DomTree::createResElement(int num)
{
    Q_ASSERT(num > 0);
    Q_ASSERT(m_datInfo != 0);

    if (num < m_bufSize && m_resStatus[ num ] != HTML_NOTPARSED) {
        // already parsed
        return true;
    }

    if (num >= m_bufSize) {
        // resize buffer size
        if (m_bufSize == 0) m_bufSize = 100;
        while (num >= m_bufSize) m_bufSize += 500;

        m_titleElm.resize(m_bufSize);
        m_bodyElm.resize(m_bufSize);
        int statusSize = m_resStatus.count();
        m_resStatus.resize(m_bufSize);
        for (int i = statusSize; i < m_bufSize; i++) {
            m_resStatus[i] = false;
        }
        int numSize = m_coloredNum.count();
        m_coloredNum.resize(m_bufSize);
        for (int i = numSize; i < m_bufSize; i++) {
            m_coloredNum[i] = false;
        }
    }

    // cleate elements
    QString titleHTML, bodyHTML;
    m_resStatus[ num ] = m_datInfo->getHTML(num, true, titleHTML, bodyHTML);

    if (m_resStatus[ num ] == HTML_NOTPARSED) {
        return false;
    }

    m_titleElm[ num ] = m_hdoc.createElement("DIV");
    m_titleElm[ num ].setAttribute("class", "res_title");
    m_titleElm[ num ].setAttribute("id", QString::number(num));
    m_titleElm[ num ].setInnerHTML(titleHTML);

    m_bodyElm[ num ] = m_hdoc.createElement("DIV");
    m_bodyElm[ num ].setAttribute("class", "res_body");
    m_bodyElm[ num ].setAttribute("id", QString::number(num));
    m_bodyElm[ num ].setInnerHTML(bodyHTML);

    return true;
}

/*
 * append the response
 */
bool DomTree::appendRes(int num)
{
    if (!createResElement(num)) return false;

    m_hdoc.body().appendChild(m_titleElm[ num ]);
    m_hdoc.body().appendChild(m_bodyElm[ num ]);

    if (num > m_bottomNum) m_bottomNum = num;

    return true;
}

/*
 * redraw all
 */
void DomTree::redraw(bool force)
{
    Q_ASSERT(m_datInfo != 0);

    int readNum = m_datInfo->getReadNum();

    // don't forget to reset abone here...
    m_datInfo->resetAbone();

    for (int i = 1; i <= readNum; i++) {
        QString titleHTML, bodyHTML;

        int oldStatus = m_resStatus[ i ];
        m_resStatus[ i ] = m_datInfo->getHTML(i , true, titleHTML, bodyHTML);

        if (force || oldStatus != m_resStatus[ i ]) {
            m_titleElm[ i ].setInnerHTML(titleHTML);
            m_bodyElm[ i ].setInnerHTML(bodyHTML);
        }
    }
}


/*
 * change color of number of the res which is responsed.
 *
 * See also DomTree::changeColorOfNumber() ,
 * DatInfo::copyOneLineToResDat(),
 * and DatInfo::collectResponsedRes().
 */
void DomTree::changeColorOfAllResponsedNumber()
{
    for (int i = 1; i <= m_bottomNum; ++i) {
        if (m_datInfo->isResponsed(i)) {
            changeColorOfNumber(i);
        }
    }
}

int DomTree::getBottomResNumber() const
{
    return m_bottomNum;
}

/*
 * append footer & header
 */
void DomTree::appendFooterAndHeader()
{
    Q_ASSERT(m_datInfo != 0);

    int readNum = m_datInfo->getReadNum();
    if (!readNum) return ;

    updateHeader(m_header);
    updateFooter(m_footer);

    m_hdoc.body().insertBefore(m_header, m_hdoc.body().firstChild());
    m_hdoc.body().appendChild(m_footer);
}

/*
 * append kokomadeyonda
 */
void DomTree::appendKokoyon()
{
    Q_ASSERT(m_datInfo != 0);

    int readNum = m_datInfo->getReadNum();
    if (!readNum) return ;

    int viewPos = m_datInfo->getViewPos();
    if (viewPos == 0) return ;

    int i = viewPos + 1;

    if (i <= readNum) m_hdoc.body().insertBefore(m_kokoyon, m_titleElm[ i ]);
    else m_hdoc.body().appendChild(m_kokoyon);
}

/*
 * private functions
 */

/*
 * append "A" Node to rootnode
 */
void DomTree::appendAnchorNode(const DOM::Element& root, const QString& href, const QString& linkstr)
{
    DOM::Element rootnode = root;
    DOM::Element element;

    element = rootnode.appendChild(m_hdoc.createElement("A"));
    {
        element.setAttribute("href", href);
        element.appendChild(m_hdoc.createTextNode(linkstr));
    }
}


/*
 * update header
 *
 * example
 * before: #KokomadeYonda 1- 101- 201- #ToSaigo<br><br>
 * after : #KokomadeYonda 1- 101- 201- 301- 401- #ToSaigo<br><br>
 */
void DomTree::updateHeader(DOM::Element& headerElement)
{
    if (! m_datInfo) return ;

    DOM::Element backupElement1, backupElement2, backupElement3;
    int readNum = m_datInfo->getReadNum();

    // remove <a href="#tosaigo"> and <BR>
    backupElement1 = headerElement.removeChild(headerElement.lastChild()); // BR
    backupElement2 = headerElement.removeChild(headerElement.lastChild()); // BR
    backupElement3 = headerElement.removeChild(headerElement.lastChild()); // "#tosaigo"

    DOM::Node node = headerElement.firstChild(); // node is now "#kokomade_yonda"
    node = node.nextSibling(); // " "
    node = node.nextSibling();

    // '1-', '101-' などのリンクを作成
    for (int num = 1; num < readNum ; num += 100) {
        if (node == 0) {
            QString href = QString("#%1").arg(num);
            QString linkStr = QString("%1-").arg(num);

            appendAnchorNode(headerElement, href, linkStr);
            node = headerElement.appendChild(m_hdoc.createTextNode(" "));
            node = node.nextSibling();
        } else {
            // 既にリンクが作られている場合は飛ばす
            node = node.nextSibling();
            if (node != 0) node = node.nextSibling();
        }
    }

    // restore <a href="#tosaigo"> and <BR>
    headerElement.appendChild(backupElement3); // "#tosaigo"
    headerElement.appendChild(backupElement2); // BR
    headerElement.appendChild(backupElement1); // BR
}

/* 
 * update footer
 *
 * example
 * before: #KokomadeYonda 1- 101- 201- #ToSaigo
 * after : #KokomadeYonda 1- 101- 201- 301- 401- #ToSaigo
 */
void DomTree::updateFooter(DOM::Element& footerElement)
{
    Q_ASSERT(m_datInfo != 0);

    DOM::Element backupElement;
    int readNum = m_datInfo->getReadNum();

    backupElement = footerElement.removeChild(footerElement.lastChild()); // "#tosaigo"

    DOM::Node node = footerElement.firstChild(); // node is now "#kokomade_yonda"
    node = node.nextSibling(); // " "
    node = node.nextSibling();

    // '1-', '101-' などのリンクを作成
    for (int num = 1; num < readNum ; num += 100) {
        if (node == 0) {
            QString href = QString("#%1").arg(num);
            QString linkStr = QString("%1-").arg(num);

            appendAnchorNode(footerElement, href, linkStr);
            node = footerElement.appendChild(m_hdoc.createTextNode(" "));
            node = node.nextSibling();
        } else {
            // 既にリンクが作られている場合は飛ばす
            node = node.nextSibling();
            if (node != 0) node = node.nextSibling();
        }
    }

    footerElement.appendChild(backupElement); // "#tosaigo"
}

/*
 * create header node
 */
void DomTree::createHeader()
{
    QString str;
    DOM::Element rootnode;

    rootnode = m_hdoc.createElement("DIV");
    {
        rootnode.setAttribute("kita_type", "header");
        rootnode.setAttribute("id", "header");

        str = QString::fromUtf8(KITAUTF8_KOKOYON);
        appendAnchorNode(rootnode, "#kokomade_yonda", str);
        rootnode.appendChild(m_hdoc.createTextNode(" "));

        str = QString::fromUtf8(KITAUTF8_SAIGO);
        appendAnchorNode(rootnode, "#tosaigo", str);

        rootnode.appendChild(m_hdoc.createElement("BR"));
        rootnode.appendChild(m_hdoc.createElement("BR"));
    }

    m_header = rootnode;
}


/*
 * create footer node
 */
void DomTree::createFooter()
{
    QString str;
    DOM::Element rootnode;

    rootnode = m_hdoc.createElement("DIV");
    {
        rootnode.setAttribute("kita_type", "footer");
        rootnode.setAttribute("id", "footer");

        str = QString::fromUtf8(KITAUTF8_KOKOYON);
        appendAnchorNode(rootnode, "#kokomade_yonda", str);
        rootnode.appendChild(m_hdoc.createTextNode(" "));

        str = QString::fromUtf8(KITAUTF8_SAIGO);
        appendAnchorNode(rootnode, "#tosaigo", str);
    }

    m_footer = rootnode;
}

/*
 * create kokomadeyonda node
 */
void DomTree::createKokoyon()
{
    QString str, style;
    DOM::Element rootnode;

    str = QString::fromUtf8(KITAUTF8_KOKOYON2);

    rootnode = m_hdoc.createElement("DIV");
    {
        rootnode.setAttribute("class", "kokoyon");
        rootnode.setAttribute("kita_type", "kokoyon");
        rootnode.setAttribute("id", "kokomade_yonda");
        rootnode.appendChild(m_hdoc.createTextNode(str));
    }

    m_kokoyon = rootnode;
}

/*
 * change color of number
 * specify color like this:  "a.coloredLink:link{ color: red; }"
 */
void DomTree::changeColorOfNumber(int num)
{
    if (m_coloredNum[ num ]) return ;

    m_coloredNum[ num ] = true;

    DOM::Node node = m_titleElm[ num ];
    node = node.firstChild();

    static_cast<DOM::HTMLElement>(node).setAttribute("class", "coloredLink");
}
