/***************************************************************************
*   Copyright (C) 2004 by Kita Developers                                 *
*   ikemo@users.sourceforge.jp                                            *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
***************************************************************************/

#include "htmlpart.h"

#include <QtCore/QRegExp>
#include <QtGui/QApplication>
#include <QtGui/QClipboard>
#include <QtGui/QCursor>
#include <QtGui/QScrollBar>
#include <QtGui/QStyle>
#include <QtGui/QStyleOptionComplex>

#include <kaction.h>
#include <khtml_events.h>
#include <kmenu.h>
#include <kmessagebox.h>
#include <krun.h>

#include "const.h"
#include "domtree.h"
#include "respopup.h"
#include "viewmediator.h"
#include "kitaui/htmlview.h"
#include "libkita/abone.h"
#include "libkita/boarddatabase.h"
#include "libkita/colorconfig.h"
#include "libkita/datmanager.h"
#include "libkita/globalconfig.h"
#include "libkita/kita_misc.h"

using namespace Kita;
/*-------------------------------------------*/


/*-------------------------------------*/
/* Don't forget to call setup() later. */

HTMLPart::HTMLPart(QWidget* parent) : KHTMLPart(new HTMLView(this, parent))
{
    m_mode = HTMLPART_MODE_MAINPART;
    m_popup = 0;
    m_domtree = 0;
    m_datUrl.clear();
    m_updatedKokoyon = false;

    clearPart();
    createHTMLDocument();
    connectSignals();
}


HTMLPart::~HTMLPart()
{
    clearPart();
}


/* private */
void HTMLPart::clearPart()
{
    slotDeletePopup();

    // delete DomTree
    delete m_domtree;
    m_domtree = 0;

    DatManager datManager(m_datUrl);
    // update ViewPos
    if (m_mode == HTMLPART_MODE_MAINPART && !m_updatedKokoyon
            && !m_datUrl.isEmpty()) {
        int readNum = datManager.getReadNum();
        if (readNum) {
            datManager.setViewPos(readNum);
        }
    }
    m_updatedKokoyon = false;

    // clear variables
    m_anchorStack.clear();
    m_centerNum = 0;
    m_jumpNumAfterLoading = 0;
    findTextInit();

    if (!m_datUrl.isEmpty()) {  // This part is opened.
        // This part is on the main thread view.
        if (m_mode == HTMLPART_MODE_MAINPART) {

            // tell Thread class that "thread is closed"
            datManager.setMainThreadOpened(false);

            // emit "deactivated all thread view" SIGNAL
            KUrl nullUrl("");
            ViewMediator::getInstance()->changeWriteTab(nullUrl);

            //  update subject tab.
        }
    }

    m_datUrl.clear();
    m_mode = HTMLPART_MODE_MAINPART;
}



/* public */
bool HTMLPart::setup(int mode, const KUrl& url)
{
    Q_ASSERT(!url.isEmpty());

    clearPart();

    m_datUrl = getDatUrl(url);
    m_mode = mode;

    if (m_mode == HTMLPART_MODE_MAINPART) {
        // This part is on the main thread view.
        DatManager datManager(m_datUrl);
        // create DatInfo explicitly to open new thread.
        // Usually, DatInfo is NOT created if ReadNum == 0.
        // See also DatManager::createDatInfo() and
        //          DatManager::getDatInfo().
        datManager.createDatInfo();

        // tell Thread class that "thread is opened"
        datManager.setMainThreadOpened(true);

        // reset abone
        datManager.resetAbone();
    }

    // create HTML Document
    createHTMLDocument();

    // create DOM manager
    if (m_mode == HTMLPART_MODE_MAINPART) {
        m_domtree = new DomTree(htmlDocument(), m_datUrl);
    }

    return true;
}


/* private */
void HTMLPart::connectSignals()
{
    // popup
    connect(this, SIGNAL(onURL(const QString&)), SLOT(slotOnUrl(const QString&)));

    connect(view(), SIGNAL(leave()), SLOT(slotLeave()));
    connect(view()->verticalScrollBar(), SIGNAL(sliderReleased()), SLOT(slotVSliderReleased()));
    connect(view()->horizontalScrollBar(), SIGNAL(sliderReleased()), SLOT(slotHSliderReleased()));
}



/* private */
void HTMLPart::createHTMLDocument()
{
    // style
    QString style = QString("body { font-size: %1pt; font-family: \"%2\"; color: %3; background-color: %4; }")
                    .arg(GlobalConfig::threadFont().pointSize())
                    .arg(GlobalConfig::threadFont().family())
                    .arg(ColorConfig::thread().name())
                    .arg(ColorConfig::threadBackground().name());

    QString text = "<html><head><style>";
    text += DEFAULT_STYLESHEET;
    text += style;
    text += "</style></head><body></body></html>";

    setJScriptEnabled(false);
    setJavaEnabled(false);

    // Use dummy URL here, and protocol should be "file:".
    // If protocol is "http:", local image files are not shown
    // (for security reasons?).
    begin(KUrl("file:/dummy.htm"));
    write(text);
    end();
}

/*---------------------------------------------------------------*/
/*---------------------------------------------------------------*/
/* direct rendering functions */

/**
 * show responses.
 * @warning don't forget to call updateScreen() later.
 */
void HTMLPart::showResponses(int startnum, int endnum)
{
    if (!m_domtree) return ;

    for (int i = startnum ; i <= endnum; i++) m_domtree->appendRes(i);
}


/* do parsing only.            */
/* call showResponses() later  */ /* public */
void HTMLPart::parseResponses(int startnum, int endnum)
{
    if (!m_domtree) return ;

    for (int i = startnum ; i <= endnum; i++) m_domtree->createResElement(i);
}


/*------------------------------------------------*/
/* Show all responses ,header, footer, and etc,
   if these are not shown.                        */

/* note that updateScreen() is called internally. */
/* So, you need not call it later.                */  /* public slot */
void HTMLPart::showAll()
{
    if (!m_domtree) return ;

    int bottom = m_domtree->getBottomResNumber();
    int readNum = DatManager(m_datUrl).getReadNum();
    if (bottom != readNum) {

        QCursor qc; qc.setShape(Qt::WaitCursor);
        QApplication::setOverrideCursor(qc);

        showResponses(1, readNum);
        updateScreen(true, false);

        QApplication::restoreOverrideCursor();
    }
}

/*
 * update screen
 */
void HTMLPart::updateScreen(bool showHeaderEtc, bool clock)
{
    if (!m_domtree) {
        view()->setFocus();
        return ;
    }

    // show clock cursor
    if (clock) {
        QCursor qc; qc.setShape(Qt::WaitCursor);
        QApplication::setOverrideCursor(qc);
    }

    // show header, footer, and kokomadeyonda, etc.
    if (showHeaderEtc) {
        m_domtree->appendKokoyon();
        m_domtree->appendFooterAndHeader();
    }

    // change color of number of the res which is responsed.
    if (m_mode == HTMLPART_MODE_MAINPART) {

        if (GlobalConfig::checkResponsed()) {
            m_domtree->changeColorOfAllResponsedNumber();
        }
    }

    // update display
    htmlDocument().applyChanges();
    view()->layout();
    view()->setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOn);
    view()->setFocus();

    // restore cursor
    if (clock) {
        QApplication::restoreOverrideCursor();
    }
}


/* public */
void HTMLPart::setInnerHTML(const QString& innerHTML)
{
    createHTMLDocument();
    htmlDocument().body().setInnerHTML(innerHTML);
}



/* redraw screen  */
void HTMLPart::redrawHTMLPart(const KUrl& datUrl, bool force)
{
    if (m_domtree == 0) return ;
    if (m_datUrl != datUrl) return ;

    m_domtree->redraw(force);
}

/* public slot */
void HTMLPart::slotSetFaceOfHTMLPart()
{
    QFont font = GlobalConfig::threadFont();

    DOM::CSSStyleDeclaration style = htmlDocument().body().style();
    style.setProperty("font-family", font.family(), "");
    style.setProperty("font-size", QString("%1pt").arg(font.pointSize()), "");
    style.setProperty("color", ColorConfig::thread().name(), "");
    style.setProperty("background-color",
            ColorConfig::threadBackground().name(), "");
    htmlDocument().applyChanges();
}

/*---------------------------------------------------------------*/
/*---------------------------------------------------------------*/
/* cache functions */


/*----------------------------------*/
/* load thread from cache & show it */ /* public */

/*  top = centerNum - preShowNum
    bottom = centerNum + afterShowNum
    readNum = DatManager::getReadNum
 
    No.1 <- show -> No.20 <- not show -> No.(top) <- show -> No.(bottom) <- not show -> No.(readNum) */

bool HTMLPart::load(int centerNum)
{
    m_centerNum = centerNum;
    m_jumpNumAfterLoading = 0;

    if (m_mode != HTMLPART_MODE_MAINPART)
        return false;
    if (!m_domtree)
        return false;
    DatManager datManager(m_datUrl);
    if (datManager.getReadNum() == 0)
        return false;

    int endNum = datManager.getReadNum();
    showResponses(1, endNum);
    updateScreen(true , false);
    gotoAnchor(QString::number(m_centerNum), false);
    view() ->setFocus();

    return true;
}



/*----------------------------*/
/* start reloading            */

/* see also slotReceiveData()
   and slotFinishLoad().      */ /* public */
bool HTMLPart::reload(int jumpNum)
{
    if (!m_domtree) return false;
    if (m_mode != HTMLPART_MODE_MAINPART) {
        // If this is not MainPart, then open MainPart.
        ViewMediator::getInstance()->openUrl(m_datUrl);
        return false;
    }

    m_firstReceive = true;
    if (m_centerNum == 0) m_centerNum = m_domtree->getBottomResNumber();
    m_jumpNumAfterLoading = jumpNum;

    // DatManager will call back slotReceiveData and slotFinishLoad.
    DatManager(m_datUrl).updateCache(this);
    view() ->setFocus();

    return true;
}



/*---------------------------------------------*/
/* This slot is called after DatManager
   received new data, and emits receiveData()  */ /* !!! "public" slot !!! */
void HTMLPart::slotReceiveData()
{
    const int delta = 20;

    if (m_mode != HTMLPART_MODE_MAINPART) return ;
    if (!m_domtree) return ;

    int readNum = DatManager(m_datUrl).getReadNum();
    int bottom = m_domtree->getBottomResNumber();

    // parsing
    parseResponses(bottom + 1, readNum);

    // rendering
    if (m_firstReceive || bottom + delta < readNum) {
        showResponses(bottom + 1, readNum);
        updateScreen(true, false);
    }

    if (m_firstReceive && m_centerNum < readNum) {
        gotoAnchor(QString::number(m_centerNum), false);
        m_firstReceive = false;
    }

    emit receiveData();
}


/*--------------------------------------------*/
/* This slot is called after DatManager
   finished updating new data,
   and emits finishReload()                   */ /* !!! "public" slot !!! */
void HTMLPart::slotFinishLoad()
{
    if (m_mode != HTMLPART_MODE_MAINPART) return ;
    if (!m_domtree) return ;

    int bottom = m_domtree->getBottomResNumber();
    int shownNum = m_centerNum + 5000;

    showResponses(bottom + 1, shownNum);
    updateScreen(true, false);
    //    m_domtree->parseAllRes();
    m_centerNum = 0;

    if (m_jumpNumAfterLoading)
        gotoAnchor(QString::number(m_jumpNumAfterLoading), false);
    m_jumpNumAfterLoading = 0;

    emit finishReload();
}




/*---------------------------------------------------------------*/
/*---------------------------------------------------------------*/
/* goto anchor */


/* public */
bool HTMLPart::gotoAnchor(const QString& anc, bool pushPosition)
{
    if (anc.isEmpty()) return false;
    if (!m_domtree || m_mode == HTMLPART_MODE_POPUP)
        return KHTMLPart::gotoAnchor(anc);

    hidePopup();

    QString ancstr = anc;
    int res = ancstr.toInt();

    if (res > 1) {

        // is target valid?
        if (!DatManager(m_datUrl).isResValid(res))
            return false;

        ancstr = QString::number(res);
    }

    if (res == 1) ancstr = "header";
    if (pushPosition) pushCurrentPosition();

    // KHTMLPart::gotoAnchor() will fail if the thread is not shown.
    // So KHTMLPart::gotoAnchor() should be called via custom event.
    // See also HTMLPart::customEvent()
    GotoAnchorEvent * e = new GotoAnchorEvent(ancstr);
    QApplication::postEvent(this, e);  // Qt will delete it when done

    return true;
}



/* jump to kokomade yonda */ /* public slot */
void HTMLPart::slotGotoKokoyon()
{
    if (!m_domtree) return ;
    if (m_mode != HTMLPART_MODE_MAINPART) return ;

    int kokoyon = DatManager(m_datUrl).getViewPos();
    gotoAnchor(QString::number(kokoyon), false);
}



/* public slot  */
void HTMLPart::slotGobackAnchor()
{
    if (m_anchorStack.isEmpty()) return ;

    QString anc = m_anchorStack.last();
    m_anchorStack.pop_back();
    gotoAnchor(anc , false);
}




/* private */
void HTMLPart::pushCurrentPosition()
{
    m_anchorStack += getCurrentIdOfNode();
}


/* find the id of current node */ /* private */
QString HTMLPart::getCurrentIdOfNode()
{
    DOM::Node node;
    node = nodeUnderMouse();
    while (node != 0 && node.nodeName().string() != "div") node = node.parentNode();
    if (node == 0) return QString();

    return static_cast<DOM::Element>(node).getAttribute("id").string();
}

/* public slot */
void HTMLPart::slotClickGotoFooter()
{
    if (!m_domtree || m_mode != HTMLPART_MODE_MAINPART) {
        gotoAnchor("footer", false);
        return ;
    }

    int bottom = m_domtree->getBottomResNumber();
    int readNum = DatManager(m_datUrl).getReadNum();

    if (readNum != bottom) {
        showResponses(bottom + 1, readNum);
        updateScreen(true, true);
    }

    gotoAnchor("footer", false);
}


/*---------------------------------------------------------------*/
/*---------------------------------------------------------------*/
/* search */


/* private */
void HTMLPart::findTextInit()
{
    m_findNode = 0;
    m_findPos = -1;
    m_find_y = 0;
}



/* public */
bool HTMLPart::findText(const QString &query, bool reverse)
{
    if (m_mode != HTMLPART_MODE_MAINPART) return false;

    QRegExp regexp(query);
    regexp.setCaseSensitivity(Qt::CaseInsensitive);

    // init
    if (m_findNode.isNull()) {

        m_findNode = htmlDocument().body();
        m_find_y = 0;

        // move to the last child node
        if (reverse) {
            while (!m_findNode.lastChild().isNull()) m_findNode = m_findNode.lastChild();
            m_find_y = view() ->contentsHeight();
        }
    }

    while (1) {

        if (m_findNode.nodeType() == DOM::Node::TEXT_NODE
                || m_findNode.nodeType() == DOM::Node::CDATA_SECTION_NODE) {

            // find the word in the current node
            DOM::DOMString nodeText = m_findNode.nodeValue();
            QString nodestr = nodeText.string();
            if (reverse && m_findPos != -1) nodestr.resize(m_findPos);

            if (reverse) m_findPos = nodestr.lastIndexOf(regexp, m_findPos);
            else m_findPos = nodestr.indexOf(regexp, m_findPos + 1);

            // scroll & select & return
            if (m_findPos != -1) {
                int matchLen = regexp.matchedLength();

                QRect qr = m_findNode.getRect();
                view() ->setContentsPos(qr.left() - 50, m_find_y - 100);
                DOM::Range rg(m_findNode, m_findPos, m_findNode, m_findPos + matchLen);
                setSelection(rg);

                return true;
            }

        } else if (m_findNode.nodeName().string() == "table") {

            QRect qr = m_findNode.getRect();

            m_find_y = qr.bottom();

        } else if (m_findNode.nodeName().string() == "div") {

            QRect qr = m_findNode.getRect();

            if (reverse) {
                m_find_y = qr.bottom();
            } else {
                m_find_y = qr.top();
            }

        } else if (m_findNode.nodeName().string() == "br") {

            DOM::Node tmpnode = m_findNode.previousSibling();

            if (tmpnode != 0) {

                QRect qr = tmpnode.getRect();
                if (reverse) m_find_y -= qr.bottom() - qr.top();
                else m_find_y += qr.bottom() - qr.top();
            }
        }

        //------------------------

        m_findPos = -1;
        DOM::Node next;

        // move to the next node
        if (!reverse) {

            next = m_findNode.firstChild();
            if (next.isNull()) next = m_findNode.nextSibling();

            while (!m_findNode.isNull() && next.isNull()) {
                m_findNode = m_findNode.parentNode();
                if (!m_findNode.isNull()) {
                    next = m_findNode.nextSibling();
                }
            }
        }
        // revearse
        else {

            next = m_findNode.lastChild();
            if (next.isNull()) next = m_findNode.previousSibling();

            while (!m_findNode.isNull() && next.isNull()) {
                m_findNode = m_findNode.parentNode();
                if (!m_findNode.isNull()) {
                    next = m_findNode.previousSibling();
                }
            }
        }

        m_findNode = next;
        if (m_findNode.isNull()) {
            m_findNode = 0;
            return false;
        }
    }

    return false;
}






/*---------------------------------------------------------------*/
/*---------------------------------------------------------------*/
/* popup menu */


/* private */
void HTMLPart::showPopupMenu(const KUrl& kurl)
{
    QString url = kurl.prettyUrl();
    bool showppm = false;

    QString str;

    // If selected Text is composed of only digits, then show res popup.
    if (!m_pushctrl && showSelectedDigitPopup()) return ;

    //-----------------------------------
    // create menu items                 

    KMenu popupMenu(view());
    KMenu* backSubMenu = 0;
    KMenu* markSubMenu = 0;

    //------
    // jump
    KAction* homeLinkAct = 0;
    KAction* kokoLinkAct = 0;
    KAction* endLinkAct = 0;
    if (m_domtree &&
            (m_mode == HTMLPART_MODE_MAINPART)) {

        showppm = true;

        DatManager datManager(m_datUrl);
        // back
        if (!m_anchorStack.isEmpty()) {
            backSubMenu = new KMenu(view());

            int i = m_anchorStack.size();
            QStringList::iterator it;
            for (it = m_anchorStack.begin(); it != m_anchorStack.end(); it++, i--) {
                str = (*it) + "   " + datManager.getPlainBody((*it).toInt()).left(10);
                KAction* backLink = new KAction(str, this);
                backSubMenu->addAction(backLink);
            }
            backSubMenu->setTitle(
                    i18nc("@title:menu Go back to the responses", "Back"));
            popupMenu.addMenu(backSubMenu);
            popupMenu.addSeparator();
        }

        // mark
        int readNum = datManager.getReadNum();
        for (int i = 1; i <= readNum ; i++) {
            if (datManager.isMarked(i)) {
                if (!markSubMenu) {
                    markSubMenu = new KMenu(view());
                    markSubMenu->setTitle(i18nc("@title:menu", "Mark"));
                    popupMenu.addMenu(markSubMenu);
                    popupMenu.addSeparator();
                }
                str = QString::number(i) + "   "
                    + datManager.getPlainBody(i).left(10);
                KAction *gotoMarkAct = new KAction(str, this);
                markSubMenu->addAction(gotoMarkAct);
            }
        }

        // home
        homeLinkAct = new KAction(
                i18nc("@action:inmenu Go to the original post", "Start"), this);
        popupMenu.addAction(homeLinkAct);

        // template
        if (m_mode == HTMLPART_MODE_MAINPART) {
            int kokoyon = datManager.getViewPos();
            if (kokoyon) {
                str = i18nc("@action:inmenu", "Kokomade Yonda (%1)", kokoyon);
                kokoLinkAct = new KAction(str, this);
                popupMenu.addAction(kokoLinkAct);
            }
        }
        // end 
        endLinkAct = new KAction(i18nc("@action:inmenu", "End"), this);
        popupMenu.addAction(endLinkAct);
    }

    // copy & abone
    KAction* copyStrAct = 0;
    KAction* aboneWordAct = 0;
    if (hasSelection()) {
        if (showppm) popupMenu.addSeparator();
        showppm = true;

        copyStrAct = new KAction(i18nc("@action:inmenu", "Copy"), this);
        popupMenu.addAction(copyStrAct);

        // truncated
        QString selectedStr = selectedText();
        if (selectedStr.length() > 20) {
            selectedStr.truncate(20);
            selectedStr.append("...");
        }

        aboneWordAct = new KAction(i18nc("@action:inmenu",
                    "Add '%1' to abone list", selectedStr), this);
        popupMenu.addAction(aboneWordAct);
    }


    // copy link
    KAction* openBrowserAct = 0;
    KAction* copyLinkAct = 0;
    if (!url.isEmpty()) {
        if (showppm) popupMenu.addSeparator();
        showppm = true;

        openBrowserAct= new KAction(i18nc("@action:inmenu",
                    "Open with Web Browser"), this);
        popupMenu.addAction(openBrowserAct);
        popupMenu.addSeparator();

        copyLinkAct = new KAction(i18nc("@action:inmenu",
                    "Copy Link Location"), this);
        popupMenu.addAction(copyLinkAct);
    }


    // show menu
    if (showppm) {

        QClipboard * clipboard = QApplication::clipboard();

        QAction* action = popupMenu.exec(QCursor::pos());
        if (!action) {
            delete backSubMenu;
            delete markSubMenu;
            return;
        }
        if (action == copyLinkAct) {
            clipboard->setText(url , QClipboard::Clipboard);
            clipboard->setText(url , QClipboard::Selection);
        } else if (action == openBrowserAct) {
            KRun::runUrl(kurl, "text/html", view());
        } else if (action == homeLinkAct) {
            gotoAnchor("header", false);
        } else if (action == kokoLinkAct) {
            slotGotoKokoyon();
        } else if (action == endLinkAct) {
            slotClickGotoFooter();
        } else if (action == copyStrAct) {
            clipboard->setText(selectedText(), QClipboard::Clipboard);
        } else if (action == aboneWordAct) {
            if (KMessageBox::warningYesNo(view(),
                    i18n("Do you want to add '%1' to abone list?",
                               selectedText())) == KMessageBox::Yes) {
                AboneConfig::aboneWordList().append(selectedText());
                redrawHTMLPart(m_datUrl, false);
            }
        } else {
            QMenu* menu = action->menu();
            if (!menu) {
                delete backSubMenu;
                delete markSubMenu;
                return;
            }
            if (menu == markSubMenu) {
                gotoAnchor(QString::number(menu->actions().indexOf(action)),
                        false);
            } else if (menu == backSubMenu) {
                for (int i = 0; i < menu->actions().indexOf(action); i++)
                    m_anchorStack.pop_back();
                slotGobackAnchor();
            }
        }
    }
    delete backSubMenu;
    delete markSubMenu;
}

/*---------------------------------------------------------------*/
/*---------------------------------------------------------------*/
/* user event */


/* protected */ /* virtual */
void HTMLPart::customEvent(QEvent * e)
{
    if (e->type() == EVENT_GotoAnchor) {
        KHTMLPart::gotoAnchor(static_cast< GotoAnchorEvent* >(e) ->getAnc());
        return ;
    }

    KHTMLPart::customEvent(e);
}


/*---------------------------------------------------------------*/
/*---------------------------------------------------------------*/
/* mouse event */


/* protected */
void HTMLPart::khtmlMousePressEvent(khtml::MousePressEvent* e)
{
    emit mousePressed(); // to ThreadView to focus this view.

    KUrl kurl;
    if (!e->url().string().isEmpty()) {
        BoardDatabase db(m_datUrl);
        kurl = KUrl(db.boardUrl(), e->url().string());
    }

    m_pushctrl = m_pushmidbt = m_pushrightbt = false;
    if (e->qmouseEvent() ->button() & Qt::RightButton) m_pushrightbt = true;
    if (e->qmouseEvent() ->modifiers() & Qt::ControlModifier) m_pushctrl = true;
    if (e->qmouseEvent() ->button() & Qt::MidButton) m_pushmidbt = true;

    if (e->url() != 0) {

        if (e->url().string().at(0) == '#') { // anchor
            kurl = m_datUrl;
            kurl.setRef(e->url().string().mid(1)) ;
        }

        clickAnchor(kurl);
        m_pushctrl = m_pushmidbt = m_pushrightbt = false;
        return ;
    }

    // popup menu
    if (m_pushrightbt) {
        showPopupMenu(kurl);
        m_pushctrl = m_pushmidbt = m_pushrightbt = false;
        return ;
    }

    KHTMLPart::khtmlMousePressEvent(e);
}





/*-------------------------------------------------------*/
/*-------------------------------------------------------*/
/* click */



/* private slot */
void HTMLPart::slotOpenUrlRequest(const KUrl& urlin, const KParts::OpenUrlArguments&)
{
    clickAnchor(urlin);
}


/*------------------------------------------------------*/
/* This function is called when user clicked res anchor */ /* private */
void HTMLPart::clickAnchor(const KUrl& urlin)
{
    QString refstr;
    KUrl datUrl = getDatUrl(urlin , refstr);

    //--------------------
    // Ctrl + right click
    if (m_pushctrl && m_pushrightbt) {
        showPopupMenu(urlin);
        return ;
    }

    //--------------------------------
    // If this is not anchor, then
    // emit openURLRequest and return

    if (datUrl.host() != m_datUrl.host() || datUrl.path() != m_datUrl.path()) {

        // right click
        if (m_pushrightbt) {

            // start multi-popup mode or show popup menu
            if (!startMultiPopup()) showPopupMenu(urlin);

            return ;
        }
        // right click

        ViewMediator::getInstance()->openUrl(urlin);
        return ;
    }

    if (refstr.isEmpty()) return ;

    //---------------------------
    // show popupmenu for #write

    if (refstr.left(5) == "write") {
        showWritePopupMenu(refstr);
        return ;
    }

    //----------------------------
    // extract responses by ID

    if (refstr.left(5) == "idpop") {
        showIdPopup(refstr);
        return ;
    }

    //---------------------------
    // show popupmenu for #bepop

    if (refstr.left(5) == "bepop") {
        showBePopupMenu(refstr);
        return ;
    }

    //-------------------------
    // start multi-popup mdde
    if (m_pushrightbt && startMultiPopup()) return ;


    //----------------------------
    // next 100 ,before 100 ,etc.
    if (m_mode == HTMLPART_MODE_MAINPART) {
        if (refstr.left(7) == "tosaigo") {
            slotClickGotoFooter();
            return;
        }
    }

    //-------------------------------
    // open Kita Navi or goto anchor

    int refNum, refNum2;

    int i = refstr.indexOf("-");
    if (i != -1) {
        refNum = refstr.left(i).toInt();
        refNum2 = refstr.mid(i + 1).toInt();
        if (refNum2 < refNum) {
            refNum2 = refNum;
        }
    } else {
        refNum = refNum2 = refstr.toInt();
    }

    if (!refNum) return ;

    if (m_mode == HTMLPART_MODE_POPUP) {
        ViewMediator::getInstance()->openUrl(urlin);
    } else {
        gotoAnchor(QString::number(refNum), true);
    }
}



/*---------------------------------------------------------*/
/* popup menu that is opened when user clicked res number. */
/* This funtcion is called in only clickAnchor().          */ /* private */
void HTMLPart::showWritePopupMenu(const QString& refstr)
{
    QClipboard * clipboard = QApplication::clipboard();
    QString str, resstr;
    int resNum = refstr.mid(5).toInt();
    DatManager datManager(m_datUrl);
    QString namestr = datManager.getPlainName(resNum);

    // show res tree
    if (m_pushrightbt) {
        int num;
        QString htmlstr = datManager.getTreeByRes(resNum, num);
        if (!num) return ;
        QString tmpstr = QString("<DIV>No.%1 : [%2]<BR>").arg(resNum).arg(num);
        tmpstr += htmlstr + "<BR><BR></DIV>";
        showPopup(m_datUrl, tmpstr);
        startMultiPopup();
        return ;
    }

    //---------------------
    // create popup menu
    QString plainStr;

    KMenu popupMenu(view());

    KAction* resAct = 0;
    KAction* quoteAct = 0;
    if (m_mode == HTMLPART_MODE_MAINPART) {
        resAct = new KAction(i18nc("@action:inmenu", "Write Response"), this);
        popupMenu.addAction(resAct);
        
        quoteAct = new KAction(i18nc("@action:inmenu", "Quote This"), this);
        popupMenu.addAction(quoteAct);

        popupMenu.addSeparator();
    }

    // mark
    KAction* markAct = new KAction(i18nc("@action:inmenu", "Mark"), this);
    markAct->setCheckable(true);
    markAct->setChecked(datManager.isMarked(resNum));
    popupMenu.addAction(markAct);

    popupMenu.addSeparator();

    // open
    KAction* showBrowserAct
        = new KAction(i18nc("@action:inmenu", "Open with Web Browser"), this);
    popupMenu.addAction(showBrowserAct);

    popupMenu.addSeparator();

    // util
    KAction* resTreeAct
        = new KAction(i18nc("@action:inmenu", "Res Tree"), this);
    popupMenu.addAction(resTreeAct);

    KAction* reverseResTreeAct
        = new KAction(i18nc("@action:inmenu", "Reverse Res Tree"), this);
    popupMenu.addAction(reverseResTreeAct);

    KAction* extractNameAct
        = new KAction(i18nc("@action:inmenu", "Extract by Name"), this);
    popupMenu.addAction(extractNameAct);

    popupMenu.addSeparator();

    // copy
    KAction* copyUrlAct
        = new KAction(i18nc("@action:inmenu", "Copy URL"), this);
    popupMenu.addAction(copyUrlAct);

    KAction* copyThreadNameAct
        = new KAction(i18nc("@action:inmenu", "Copy Title and URL"), this);
    popupMenu.addAction(copyThreadNameAct);

    KAction* copyAct = new KAction(i18nc("@action:inmenu", "Copy"), this);
    popupMenu.addAction(copyAct);

    // kokkoma de yonda
    KAction* setKokoYonAct = 0;
    if (m_domtree && m_mode == HTMLPART_MODE_MAINPART) {
        popupMenu.addSeparator();

        setKokoYonAct
            = new KAction(i18nc("@action:inmenu", "Set Kokomade Yonda"), this);
        popupMenu.addAction(setKokoYonAct);
    }

    // abone
    popupMenu.addSeparator();

    KAction* aboneNameAct
        = new KAction(i18nc("@action:inmenu", "Add Name to Abone List"), this);
    popupMenu.addAction(aboneNameAct);


    // show popup menu
    QAction* action = popupMenu.exec(QCursor::pos());
    if (!action) {
        return;
    }
    if (action == resAct) {
        resstr = ">>" + QString::number(resNum) + '\n';
        ViewMediator::getInstance()->showWriteView(m_datUrl, resstr);
    } else if (action == quoteAct) {
        resstr = ">>" + QString::number(resNum) + '\n'
                 + "> " + datManager.getPlainTitle(resNum) + '\n'
                 + "> " + datManager.getPlainBody(resNum).replace('\n', "\n> ") + '\n';
        ViewMediator::getInstance()->showWriteView(m_datUrl, resstr);
    } else if (action == markAct) {
        datManager.setMark(resNum, !datManager.isMarked(resNum));
    } else if (action == copyAct || action == copyUrlAct || action == copyThreadNameAct) {
        str.clear();

        // title 
        if (action == copyThreadNameAct || action == copyAct) {
            str = datManager.threadName();
        }

        // url
        if (!str.isEmpty()) str += '\n';
        str += datManager.threadUrl() + '/' + QString::number(resNum)
            + '\n';

        // body
        if (action == copyAct) {
            str += '\n'
                   + datManager.getPlainTitle(resNum) + '\n'
                   + datManager.getPlainBody(resNum) + '\n';
        }

        // copy
        clipboard->setText(str , QClipboard::Clipboard);
        clipboard->setText(str , QClipboard::Selection);

    } else if (action == setKokoYonAct) {
        datManager.setViewPos(resNum);
        ViewMediator::getInstance()->updateBoardView(m_datUrl);
        m_updatedKokoyon = true;
        updateScreen(true, true);
        gotoAnchor(QString::number(resNum), false);

    } else if (action == showBrowserAct) {
        str = datManager.threadUrl() + '/' + QString::number(resNum);

        KRun::runUrl(str, "text/html", view());
    } else if (action == aboneNameAct) {
        if (KMessageBox::warningYesNo(view(),
                i18n("Do you want to add '%1' to abone list?", namestr),
                "Kita") == KMessageBox::Yes) {
            AboneConfig::aboneNameList().append(namestr);
            redrawHTMLPart(m_datUrl, false);
        }
    }
}



/*--------------------------------------------------*/
/* popup that is opened when user clicked ID        */
/* This funtcion is called in only clickAnchor().   */ /* private */
void HTMLPart::showIdPopup(const QString& refstr)
{
    QString strid = refstr.mid(5)
                    .replace("%2B", "+")
                    .replace("%2F", "/");

    // popup
    if (m_pushrightbt) {
        int num;
        QString htmlstr = DatManager(m_datUrl).getHtmlById(strid, num);
        if (num <= 1) return ;
        QString tmpstr = QString("<DIV>ID:%1:[%2]<BR>").arg(strid).arg(num);
        tmpstr += htmlstr + "<BR><BR></DIV>";
        showPopup(m_datUrl, tmpstr);
        startMultiPopup();
    }

    else {
        KMenu popupMenu(view());

        KAction* aboneAct = new KAction(i18nc("@action:inmenu",
                    "Add Id to Abone List"), this);
        popupMenu.addAction(aboneAct);

        QAction* action = popupMenu.exec(QCursor::pos());
        if (!action) {
            return;
        }
        if (action == aboneAct) {
            // add ID to abone list
            if (KMessageBox::warningYesNo(view(),
                    i18n("Do you want to add '%1' to abone list?", strid),
                    "Kita") == KMessageBox::Yes) {
                AboneConfig::aboneIDList().append(strid);
                redrawHTMLPart(m_datUrl, false);
            }
        }
    }
}



/*---------------------------------------------------------*/
/* popup menu that is opened when user clicked be anchor.  */
/* This funtcion is called in only clickAnchor().          */ /* private */
void HTMLPart::showBePopupMenu(const QString& refstr)
{
    QString strUrl = "http://be.2ch.net/test/p.php?i=" + refstr.mid(5)
                    + "&u=d:" + DatManager(m_datUrl).threadUrl() + "/l50";

    if (m_pushrightbt) {
        // create popup menu
        KMenu popupMenu(view());
        
        QClipboard * clipboard = QApplication::clipboard();

        KAction* copyUrlAct
            = new KAction(i18nc("@action:inmenu", "copy URL"), this);
        popupMenu.addAction(copyUrlAct);

        KAction* showBrowserAct = new KAction(i18nc("@action:inmenu",
                    "Open with Web Browser"), this);

        // show popup menu
        QAction* action = popupMenu.exec(QCursor::pos());
        if (!action) {
            return;
        }
        if (action == copyUrlAct) {
            // copy
            clipboard->setText(strUrl, QClipboard::Clipboard);
            clipboard->setText(strUrl, QClipboard::Selection);
        } else if (action == showBrowserAct) {
            KRun::runUrl(strUrl, "text/html", view());
        }
    } else {
        KRun::runUrl(strUrl, "text/html", 0);
    }
}

/*-------------------------------------------------------*/
/*-------------------------------------------------------*/
/* popup */


/* public */
bool HTMLPart::isPopupVisible()
{
    if (!m_popup) return false;
    return m_popup->isVisible();
}


/* public slot */
void HTMLPart::slotDeletePopup()
{
    delete m_popup;
    m_popup = 0;
    m_multiPopup = false;
}


/* for convenience */ /* public slot */
void HTMLPart::slotShowResPopup(const QPoint& point, int refNum, int refNum2)
{
    QString innerHTML = DatManager(m_datUrl).getHtml(refNum, refNum2);
    if (innerHTML.isEmpty()) return ;

    showPopupCore(m_datUrl, innerHTML, point);
}


/* for convenience */ /* private */
void HTMLPart::showPopup(const KUrl& url, const QString& innerHTML)
{
    showPopupCore(url, innerHTML, QCursor::pos());
}


/* show popup window   */  /* private */
void HTMLPart::showPopupCore(const KUrl& url, const QString& innerHTML, const QPoint& point)
{
    slotDeletePopup();
    m_multiPopup = false;

    m_popup = new ResPopup(view() , url);

    connect(m_popup, SIGNAL(hideChildPopup()), SLOT(slotHideChildPopup()));

    m_popup->setText(innerHTML);
    m_popup->adjustSize();
    m_popup->adjustPos(point);
    m_popup->show();
}


/*------------------------*/
/* start multi-popup mode */ /* private */
bool HTMLPart::startMultiPopup()
{

    if (m_popup && m_popup->isVisible()) {
        m_multiPopup = true;
        m_popup->moveMouseAbove();
    } else {
        m_multiPopup = false;
    }

    return m_multiPopup;
}


/* Is it multi-popup mode now? */ /* private */
bool HTMLPart::isMultiPopupMode()
{
    if (!m_popup) {
        m_multiPopup = false;
    } else if (m_popup->isHidden()) {
        m_multiPopup = false;
    }

    return m_multiPopup;
}

/* private */
void HTMLPart::hidePopup()
{
    if (m_popup) {
        m_popup->hide();
    }
    m_multiPopup = false;
}

/* return true if this view is under mouse. */ /* private */
bool HTMLPart::isUnderMouse(int mrgwd, int mrght)
{
    QPoint pos = QCursor::pos();
    int cx = pos.x(), cy = pos.y();

    QPoint viewpos = view() ->mapToGlobal(QPoint(0, 0));
    int px = viewpos.x(), py = viewpos.y();
    int wd = view() ->visibleWidth(), ht = view() ->visibleHeight();

    if ((px < cx && cx < px + wd + mrgwd)
            && (py < cy && cy < py + ht + mrght)) {
        return true;
    }

    return false;
}

/* private slot */
void HTMLPart::slotLeave()
{
    if (isMultiPopupMode()) return ;
    if (view()->horizontalScrollBar()->isSliderDown()) return;
    if (view()->verticalScrollBar()->isSliderDown()) return;

    hidePopup();

    // emit signal to have parent hide this if this is popup.
    if (m_mode == HTMLPART_MODE_POPUP && !isUnderMouse(0, 0)) {
        emit hideChildPopup();
    }
}

/* private slot */
void HTMLPart::slotVSliderReleased()
{
    QScrollBar * bar = view() ->verticalScrollBar();
    QRect rt = bar->style()->subControlRect(QStyle::CC_Slider, new QStyleOptionComplex(), QStyle::SC_SliderHandle);
    int mrg = rt.right() - rt.left();

    hidePopup();

    // emit signal to have parent hide this if this is popup.
    if (m_mode == HTMLPART_MODE_POPUP && !isUnderMouse(mrg, 0)) {
        emit hideChildPopup();
    }
}


/* private slot */
void HTMLPart::slotHSliderReleased()
{
    QScrollBar * bar = view() ->horizontalScrollBar();
    QRect rt = bar->style()->subControlRect(QStyle::CC_Slider, new QStyleOptionComplex(), QStyle::SC_SliderHandle);
    int mrg = rt.bottom() - rt.top();

    hidePopup();

    // emit signal to have parent hide this if this is popup.
    if (m_mode == HTMLPART_MODE_POPUP && !isUnderMouse(0, mrg)) {
        emit hideChildPopup();
    }
}



/* private slot */
void HTMLPart::slotHideChildPopup()
{
    hidePopup();

    // emit signal to have parent hide this if this is popup.
    if (m_mode == HTMLPART_MODE_POPUP && !isUnderMouse(0, 0)) {
        emit hideChildPopup();
    }
}


/*---------------------------------------------------*/
/* This slot is called when mouse moves onto the URL */ /* private slot */
void HTMLPart::slotOnUrl(const QString& url)
{
    // config

    const int maxpopup = 10; // max number of responses shown in the popup window

    //----------------------------

    if (isMultiPopupMode()) return ;

    slotDeletePopup();

    if (url.isEmpty()) return ;
    if (url.left(7) == "mailto:") return ;

    // Is Kita active now?
    if(ViewMediator::getInstance()->isActive() == false) return;

    // get reference
    QString refstr;
    KUrl datUrl = m_datUrl;
    if (url.at(0) == '#') {
        refstr = url.mid(1);
    } else {
        datUrl = getDatUrl(KUrl(m_datUrl, url) , refstr);
    }

    DatManager datManager(m_datUrl);
    //------------------------
    // id popup

    if (url.left(6) == "#idpop") {
        int num = datManager.getNumById(url.mid(6));
        QString tmpstr;
        if (num >= 2) {
            tmpstr = QString("<DIV>ID:%1:[%2]</DIV>").arg(url.mid(6)).arg(num);
        } else {
            tmpstr = "<DIV>"
                + i18nc("@label:textbox This is the first reply", "None")
                + "</DIV>";
        }
        showPopup(m_datUrl, tmpstr);
        return ;
    }


    //------------------------
    // show reffered num

    if (refstr.left(5) == "write") {
        int no = refstr.mid(5).toInt();
        int num = 0;
        datManager.getTreeByRes(no, num);
        QString tmpstr;
        if (num) {
            tmpstr = QString("<DIV>No.%1 : [%2]</DIV>").arg(no).arg(num);
        } else {
            tmpstr = "<DIV>"
                + i18nc("@label:textbox There is no reply", "None") + "</DIV>";
        }
        showPopup(m_datUrl, tmpstr);
        return ;
    }


    //------------------------
    // abone

    if (url.left(6) == "#abone") {
        int no = url.mid(6).toInt();
        QString tmpstr = datManager.getHtml(no, no, false);
        showPopup(m_datUrl, tmpstr);
        return ;
    }

    //-------------------------
    // popup for anchor

    QString innerHTML;
    int refNum;
    int refNum2;

    int i = refstr.indexOf("-");
    if (i != -1) { // >>refNum-refNum2

        refNum = refstr.left(i).toInt();
        refNum2 = refstr.mid(i + 1).toInt();

        if (refNum) {
            if (refNum2 < refNum) refNum2 = refNum;
            if (refNum2 - refNum > maxpopup - 1) refNum2 = refNum + maxpopup - 1;
        }

    } else { // >>refNum
        refNum = refstr.toInt();
        refNum2 = refNum;
    }

    DatManager localDatManager(datUrl);
    // another thread?
    if (datUrl.host() != m_datUrl.host() || datUrl.path() != m_datUrl.path()) {

        // get board name
        BoardDatabase db(datUrl);
        QString boardName = db.boardName();
        if (!boardName.isEmpty()) innerHTML += '[' + boardName + "] ";

        // If idx file of datURL is not read, thread name cannot be obtained.
        // so, create DatInfo if cache exists, and read idx file in DatInfo::DatInfo().
        localDatManager.getDatInfoPointer();

        // get thread Name
        QString subName = localDatManager.threadName();
        if (!subName.isEmpty()) innerHTML += subName + "<br><br>";

        if (!refNum) refNum = refNum2 = 1;
    }

    // get HTML and show it
    if (!refNum) return ;
    innerHTML += localDatManager.getHtml(refNum, refNum2);

    if (!innerHTML.isEmpty())
        showPopup(datUrl, innerHTML);
}



/* If selected Text is composed of only digits,
   then show res popup.                          */ /* private */
bool HTMLPart::showSelectedDigitPopup()
{
    if (!hasSelection()) return false;

    QString linkstr;
    int refNum;
    QString selectText = selectedText();
    const QChar *chpt = selectText.unicode();
    unsigned int length = selectText.length();

    if ((refNum = stringToPositiveNum(chpt, length)) != -1) {
        QString innerHTML = DatManager(m_datUrl).getHtml(refNum, refNum);
        if (!innerHTML.isEmpty()) {
            showPopup(m_datUrl, innerHTML);
            startMultiPopup();
            return true;
        }
    }

    return false;
}
