/***************************************************************************
*   Copyright (C) 2004 by Hideki Ikemoto , (c) 2004 by 421                *
*   ikemo@users.sourceforge.jp                                            *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
***************************************************************************/

/* Basic class of Tab widget, Tab bar, and Dock widget. */

#include "tabwidgetbase.h"

#include <QtGui/QApplication>

#include <kaction.h>
#include <kactioncollection.h>
#include <klocale.h>
#include <kshortcutsdialog.h>
#include <kstandardshortcut.h>
#include <ktabbar.h>

using namespace Kita;

TabWidgetBase::TabWidgetBase(QWidget* parent) : KTabWidget(parent)
{
    connect(this, SIGNAL(currentChanged (QWidget *)),
             SLOT(slotCurrentChanged (QWidget *)));
    setupActions();

    if (parent) {
        // setup part manager
        m_manager = new KParts::PartManager(parent, this);
        m_manager->addManagedTopLevelWidget(parent);
    } else {
        m_manager = 0;
    }
}


TabWidgetBase::~TabWidgetBase()
{
    // remove parts
    if (m_manager && !(m_manager->parts().isEmpty())) {
        KParts::Part * part;
        while ((part = m_manager->parts().first()) != 0) {
            m_manager->removePart(part);
            removePage(part->widget());
            delete part;
        }
    }
    delete m_manager;
    m_manager = 0;

    // remove widgets which don't belong to parts
    QWidget* view = currentWidget();
    while (count() > 0 && view) {
        removePage(view);
        delete view;
        view = currentWidget();
    }
}

/* public slot  */
void TabWidgetBase::slotCurrentChanged(QWidget * w)
{
    if (m_manager == 0) return ;
    if (w == 0) return ;
    w->activateWindow();
    w->setFocus();

    KParts::Part* part = findPartFromWidget(w);
    if (part) {
        m_manager->setActivePart(part);
    }
}


/* close num-th tab       */
/* see also customEvent   */ /* public slot */
void TabWidgetBase::slotCloseTab(int num)
{
    CloseTabEvent * e = new CloseTabEvent(num);
    QApplication::postEvent(this, e);  // Qt will delete it when done
}


/* Calling deleteWidget in the child part will be the
   cause of crash.  So you need to call deleteWidget
   via custom event.                                   */ /* protected */ /* virtual */
void TabWidgetBase::customEvent(QEvent * e)
{
    if (e->type() == EVENT_CloseTab) {
        deleteWidget (widget(static_cast< CloseTabEvent* >(e) ->getIndex()));
    }
}


/* protected */ /* virtual */
void TabWidgetBase::deleteWidget(QWidget* w)
{
    if (w == 0) return ;

    removePage(w);
    KParts::Part* part = findPartFromWidget(w);
    if (part) m_manager->removePart(part);
    delete w;
}


/* protected */
KParts::Part* TabWidgetBase::findPartFromWidget(QWidget* w)
{
    if (w == 0) return 0;
    if (m_manager == 0) return 0;
    if (m_manager->parts().isEmpty()) return 0;

    KParts::Part *part;
    QList<KParts::Part*>::const_iterator it = m_manager->parts().begin();
    while ((part = (*it)) != 0) {
        if (part->widget() == w) return part;
        ++it;
    }

    return 0;
}

/*------------------------------------*/
/* common tab actions */


/* private */
void TabWidgetBase::setupActions()
{
    actionCollection()->associateWidget(this);

    QString str = i18nc("@action:inmenu", "Configure Shortcuts...");
    KAction* tabConfigkeysAction
        = actionCollection()->addAction("tab_configkeys");
    tabConfigkeysAction->setText(str);
    connect(tabConfigkeysAction, SIGNAL(triggered()),
            this, SLOT(slotConfigureKeys()));

    KAction* tabNexttabAction
        = actionCollection()->addAction("tab_nexttab");
    tabNexttabAction->setText(i18nc("@action:inmenu", "Activate Next Tab"));
    tabNexttabAction->setShortcut(KStandardShortcut::tabNext());
    connect(tabNexttabAction, SIGNAL(triggered()), this, SLOT(slotNextTab()));

    KAction* tabPrevtabAction = actionCollection()->addAction("tab_prevtab");
    tabPrevtabAction->setText(i18nc("@action:inmenu",
                "Activate Previous Tab"));
    tabPrevtabAction->setShortcut(KStandardShortcut::tabPrev());
    connect(tabPrevtabAction, SIGNAL(triggered()), this, SLOT(slotPrevTab()));

    KAction* tabClosetabAction = actionCollection()->addAction("tab_closetab");
    tabClosetabAction->setText(i18nc("@action:inmenu", "Close This Tab"));
    tabClosetabAction->setShortcut(KStandardShortcut::close());
    connect(tabClosetabAction, SIGNAL(triggered()),
            this, SLOT(slotCloseCurrentTab()));

    KAction* tabCloseothertabAction
        = actionCollection()->addAction("tab_closetotherab");
    tabCloseothertabAction->setText(i18nc("@action:inmenu",
                "Close Other Tabs"));
    tabCloseothertabAction->setShortcut(KStandardShortcut::close());
    connect(tabCloseothertabAction, SIGNAL(triggered()),
            this, SLOT(slotCloseOtherTab()));

    KAction* tabCloserighttabAction
        = actionCollection()->addAction("tab_closetrightab");
    tabCloserighttabAction->setText(i18nc("@action:inmenu",
                "Close Right Tabs"));
    connect(tabCloserighttabAction, SIGNAL(triggered()),
            this, SLOT(slotCloseRightTab()));

    KAction* tabCloselefttabAction
        = actionCollection()->addAction("tab_closetleftab");
    tabCloselefttabAction->setText(i18nc("@action:inmenu", "Close Left Tabs"));
    connect(tabCloselefttabAction, SIGNAL(triggered()),
            this, SLOT(slotCloseLeftTab()));

    KAction* tabClosealltabAction
        = actionCollection()->addAction("tab_closealltab");
    tabClosealltabAction->setText(i18nc("@action:inmenu", "Close All Tabs"));
    connect(tabClosealltabAction, SIGNAL(triggered()),
            this, SLOT(slotCloseAllTab()));
}



/* public slot */
void TabWidgetBase::slotConfigureKeys()
{
    QString str = "Tab Actions";
    KShortcutsDialog dlg(KShortcutsEditor::AllActions, KShortcutsEditor::LetterShortcutsAllowed, this);
    dlg.addCollection(actionCollection(), str);
    dlg.configure();
}


/* public slot */
void TabWidgetBase::slotPrevTab()
{
    int max = count();
    int curpage = currentIndex();
    if (max <= 1) return ;

    if (curpage == 0) setCurrentIndex(max - 1);
    else setCurrentIndex(curpage - 1);
}

/* public slot */
void TabWidgetBase::slotNextTab()
{
    int max = count();
    int curpage = currentIndex();
    if (max <= 1) return ;

    if (curpage == max - 1) setCurrentIndex(0);
    else setCurrentIndex(curpage + 1);
}


/* see also customEvent */ /* public slot */
void TabWidgetBase::slotCloseCurrentTab()
{
    slotCloseTab(currentIndex());
}


/* see also customEvent */ /* public slot */
void TabWidgetBase::slotCloseOtherTab(int idx)
{
    int max = count();
    if (max == 0) return ;
    if (idx == -1) idx = currentIndex();

    int i = 0;

    while (i < max && i != idx) {
        slotCloseTab(0);
        i++;
    }

    i++;
    while (i < max) {
        slotCloseTab(1);
        i++;
    }
}


/* see also customEvent */ /* public slot */
void TabWidgetBase::slotCloseRightTab(int idx)
{
    int max = count();
    if (max == 0) return ;
    if (idx == -1) idx = currentIndex();

    int i, i2;

    i = i2 = idx + 1;
    while (i < max) {
        slotCloseTab(i2);
        i++;
    }
}


/* see also customEvent */ /* public slot */
void TabWidgetBase::slotCloseLeftTab(int idx)
{
    int max = count();
    if (max == 0) return ;
    if (idx == -1) idx = currentIndex();

    int i = 0;

    while (i < max && i != idx) {
        slotCloseTab(0);
        i++;
    }
}


/* see also customEvent */ /* public slot */
void TabWidgetBase::slotCloseAllTab()
{
    int max = count();
    if (max == 0) return ;

    int i = 0;

    while (i < max) {
        slotCloseTab(0);
        i++;
    }
}
