/***************************************************************************
*   Copyright (C) 2004 by Kita Developers                                 *
*   ikemo@users.sourceforge.jp                                            *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
***************************************************************************/
#include "access.h"

#include <QtCore/QByteArray>
#include <QtCore/QDataStream>
#include <QtCore/QDateTime>
#include <QtCore/QFile>
#include <QtCore/QRegExp>

#include <kdatetime.h>
#include <kurl.h>
#include <kio/job.h>
#include <kio/jobuidelegate.h>
#include <kio/slaveconfig.h>

#include "account.h"
#include "boarddatabase.h"
#include "cache.h"
#include "config.h"
#include "flashcgi.h"
#include "jbbs.h"
#include "kita_misc.h"
#include "k2ch.h"
#include "machibbs.h"

using namespace Kita;


Access::Access(const KUrl& datUrl)
: m_datUrl(datUrl), m_currentJob(0), m_header("HTTP/1.1 200 ") /* dummy */,
    m_dataSize(0), m_readNum(0)
{
    BoardDatabase db(m_datUrl);
    m_bbstype = db.type();
}

/* read data from cache, then emit data to DatInfo. */ /* public */
void Access::getcache()
{
    QByteArray orgData;

    // get cache path.
    Cache cache(m_datUrl);
    QString cachePath = cache.getPath();
    if (cachePath.isEmpty()) {
        return;
    }

    // read cache file.
    QFile file(cachePath);
    if (file.open(QIODevice::ReadOnly)) {
        orgData += file.readAll();
        file.close();
    }

    // set data size.
    if (orgData.isEmpty()) return ;
    m_dataSize = orgData.length();

    switch (m_bbstype) {

    case Board_2ch:
    case Board_MachiBBS:  // Machi BBS's data is already parsed as 2ch dat.

        {
            QString tmpData = K2ch().datToUnicode(orgData);
            QStringList tmpList = tmpData.split('\n');
            emit receiveData(tmpList);
        }
        break;

    default:
        // convert data stream into 2ch dat.
        // and emit receiveData SIGNAL.
        emitDatLineList(orgData);
        break;
    }
}

/* write data to cache */ /* protected */
void Access::writeCacheData()
{
    if (m_invalidDataReceived) return ;
    if (m_threadData.isEmpty()) return ;

    m_dataSize += m_threadData.length();

    Cache cache(m_datUrl);
    QString cachePath = cache.getPath();
    if (!cachePath.isEmpty()) {
        QFile file(QFile::encodeName(cachePath));
        if (!file.open(QIODevice::WriteOnly))
            return;
        QDataStream out(&file);
        out << m_threadData;
    }
    m_threadData.clear(); // clear baffer

    return ;
}

/* update cache file */ /* public */
bool Access::getupdate(int readNum)
{
    // init
    m_readNum = readNum;
    m_threadData.clear();
    m_firstReceive = false;
    m_invalidDataReceived = false;
    m_lastLine.clear();

    // set URL of data
    QString getUrl;
    switch (m_bbstype) {

    case Board_MachiBBS:
        getUrl = getThreadUrl(m_datUrl);
        if (m_readNum > 0)
            getUrl += "&START=" + QString::number(m_readNum + 1);
        InitParseMachiBBS();
        break;

    case Board_JBBS:
        getUrl = getThreadUrl(m_datUrl);
        getUrl.replace("read.cgi", "rawmode.cgi");  // adhoc...
		if (m_readNum > 0)
            getUrl += '/' + QString::number(m_readNum + 1) + '-';
        break;

    default:
        getUrl = m_datUrl.prettyUrl();
    }

    // set UserAgent
    const QString useragent = QString("Monazilla/1.00 (Kita/%1)").arg(VERSION);
    KIO::SlaveConfig::self() ->setConfigData("http",
            KUrl(getUrl).host(),
            "UserAgent", useragent);

    // create new job
    KIO::TransferJob* job
        = KIO::get(getUrl, KIO::Reload, KIO::HideProgressInfo);
    m_currentJob = job;

    connect(job, SIGNAL(data(KIO::Job*, const QByteArray&)),
            SLOT(slotReceiveThreadData(KIO::Job*, const QByteArray&)));
    connect(job, SIGNAL(result(KJob*)), SLOT(slotThreadResult(KJob*)));

    // use 'HTTP-Headers' metadata.
    job->addMetaData("PropagateHttpHeader", "true");

    // resume
    if (m_bbstype != Board_MachiBBS
            && m_bbstype != Board_JBBS
            && m_dataSize > 0) {
        m_firstReceive = true; // remove first char (i.e. \n). see also slotReceiveThreadData()
        job->addMetaData("resume", QString::number(m_dataSize - 1));
        job->addMetaData("AllowCompressedPage", "false");
    }

    return true;
}

void Access::slotThreadResult(KJob* job)
{
    KIO::TransferJob *tjob = qobject_cast<KIO::TransferJob *>(job);
    if (!tjob) {
        return;
    }
    m_currentJob = 0;
    if (tjob->error()) {
        tjob->ui()->setWindow(0);
        tjob->ui()->showErrorMessage();
    } else {
        m_header = tjob->queryMetaData("HTTP-Headers");
    }

    writeCacheData();
    emit finishLoad();
}

void Access::slotReceiveThreadData(KIO::Job*, const QByteArray& data)
{
    QByteArray data_tmp(data);

    // HACK: crash when data contains '\0'.
    for (int i=0; i< data_tmp.size(); i++) {
        if (data_tmp[ i ] == '\0') data_tmp[ i ] = ' ';
    }

    if (m_bbstype == Board_MachiBBS
            || m_bbstype == Board_JBBS) {

        emitDatLineList(data_tmp);
        return ;
    }

    // check if received data is invalid (or broken).
    if ((m_dataSize > 0 && responseCode() != 206)
            || (m_firstReceive && data_tmp[ 0 ] != '\n')
            || (m_dataSize == 0 && responseCode() != 200))
        m_invalidDataReceived = true;

    if (m_invalidDataReceived) return ;

    // If this is the first call at resumption, remove LF(\n) at head.
    if (m_firstReceive) {
        data_tmp = data_tmp.mid(1);
    }
    m_firstReceive = false;

    emitDatLineList(data_tmp);
}


/* convert data stream into 2ch dat.
   and emit receiveData SIGNAL.        */  /* private */
void Access::emitDatLineList(const QByteArray& dataStream)
{
    QList<QByteArray> lineList;
    QStringList datLineList;
    if (dataStream.isEmpty()) return ;

    bool endIsLF = false;
    if (dataStream.at(dataStream.length() - 1) == '\n') endIsLF = true;

    // split the stream
    m_lastLine += dataStream;
    lineList = m_lastLine.split('\n');
    m_lastLine.clear();

    // save the last line
    if (!endIsLF) {

        QList<QByteArray>::iterator lastit = lineList.end();
        lastit--;
        if (lastit != lineList.end()) {

            m_lastLine = (*lastit);
            lineList.removeLast();
        }
    }

    // filtering

    // convert lines into 2ch dat
    int count = lineList.count();
    for (int i = 0; i < count ; ++i) {

        if (! lineList[ i ].isEmpty()) {

            QString line, line2;
            QByteArray ba;
            int nextNum = m_readNum + 1;

            // convert line
            switch (m_bbstype) {

            case Board_MachiBBS:
                {
                    MachiBBS machiBbs;
                    line = machiBbs.datToUnicode(lineList[i]);
                    line2 = ParseMachiBBSOneLine(line, nextNum);
                    ba = machiBbs.datFromUnicode(line2);
                }
                break;

            case Board_JBBS:
                {
                    JBBS jBbs;
                    line = jBbs.datToUnicode(lineList[i]);
                    line2 = ParseJBBSOneLine(line, nextNum);
                    ba = jBbs.datFromUnicode(line2);
                }
                break;

            case Board_FlashCGI:
                {
                    FlashCGI flashCgi;
                    line = flashCgi.datToUnicode(lineList[i]);
                    line2 = ParseFlashCGIOneLine(line);
                    ba = flashCgi.datFromUnicode(line2);
                }
                break;

            default:
                line = line2 = K2ch().datToUnicode(lineList[i]);
                ba = lineList[i];
            }

            if (line2.isEmpty()) continue;

            // add abone lines
            const char aboneStr[] = "abone<><><>abone<>";
            while (nextNum > m_readNum + 1) {
                datLineList += aboneStr;
                m_threadData += aboneStr + '\n';
                ++m_readNum;
            }

            // save line
            if (m_bbstype == Board_MachiBBS) m_threadData += ba + '\n';
            else m_threadData += lineList[ i ] + '\n';
            ++m_readNum;

            datLineList += line2;
        }
    }

    // call DatInfo::slotReceiveData()
    emit receiveData(datLineList);
}



void Access::killJob()
{
    if (m_currentJob) m_currentJob->kill();
}

void Access::stopJob()
{
    if (m_currentJob) m_currentJob->kill(); // emit result signal
}

int Access::serverTime()
{
    if (m_currentJob) m_header = m_currentJob->queryMetaData("HTTP-Headers");
    //if (m_header.isEmpty()) return QDateTime::currentDateTime().toTime_t();
    // parse HTTP headers
    QStringList headerList = m_header.split('\n');
    QRegExp regexp("Date: (.*)");
    QStringList dateStrList = headerList.filter(regexp);
    if (dateStrList.isEmpty() || regexp.indexIn(dateStrList[0]) == -1) {
        // 'Date' header is not found. use current time.
        return QDateTime::currentDateTime().toTime_t();
    } else {
        return KDateTime::fromString(regexp.cap(1), KDateTime::RFCDate)
            .toTime_t();
    }
}

int Access::responseCode()
{
    if (m_currentJob) m_header = m_currentJob->queryMetaData("HTTP-Headers");
    // parse HTTP headers
    QStringList headerList = m_header.split('\n');
    QRegExp regexp("HTTP/1\\.[01] ([0-9]+) .*");
    QStringList dateStrList = headerList.filter(regexp);
    if (dateStrList.isEmpty() || regexp.indexIn(dateStrList[0]) == -1) {
        // invalid response
        if (m_bbstype == Board_JBBS) return 200; // adhoc...
        return 0;
    } else {
        return regexp.cap(1).toInt();
    }
}

/* public */
int Access::dataSize() const
{
    return m_dataSize;
}

/* public */
bool Access::invalidDataReceived() const
{
    return m_invalidDataReceived;
}

//
// access offlaw.cgi
//
QString OfflawAccess::get()
{
    QString getUrl = datToOfflaw(m_datUrl.url());
    KUrl kgetUrl(getUrl);
    kgetUrl.addQueryItem("sid", Account::getSessionId());

    m_threadData.clear();
    m_invalidDataReceived = false;

    KIO::SlaveConfig::self() ->setConfigData("http",
            KUrl(getUrl).host(),
            "UserAgent",
            QString("Monazilla/1.00 (Kita/%1)").arg(VERSION));

    KIO::TransferJob* job
        = KIO::get(kgetUrl, KIO::Reload, KIO::HideProgressInfo);
    m_currentJob = job;

    connect(job, SIGNAL(data(KIO::Job*, const QByteArray&)),
             SLOT(slotReceiveThreadData(KIO::Job*, const QByteArray&)));
    connect(job, SIGNAL(result(KJob*)), SLOT(slotThreadResult(KJob*)));

    // use 'HTTP-Headers' metadata.
    job->addMetaData("PropagateHttpHeader", "true");

    return QString(); // dummy
}

void OfflawAccess::slotThreadResult(KIO::Job* job)
{
    m_currentJob = 0;
    if (job->error()) {
        job->ui()->setWindow(0);
        job->ui()->showErrorMessage();
    } else {
        m_header = job->queryMetaData("HTTP-Headers");
    }

    if (!m_invalidDataReceived && !m_threadData.isEmpty()) {
        KUrl url = m_datUrl;
        writeCacheData();
    }
    emit finishLoad();
}

void OfflawAccess::slotReceiveThreadData(KIO::Job*, const QByteArray& data)
{
    QByteArray cstr(data);

    if ((m_dataSize > 0 && responseCode() != 206)
            || (m_dataSize == 0 && responseCode() != 200)) {
        m_invalidDataReceived = true;
    }

    if (m_invalidDataReceived) return ;

    // "+OK ....../1024K\tLocation:temp/\n"
    if (m_threadData.isEmpty() && cstr[ 0 ] == '+') {
        // skip first line.
        int index = cstr.indexOf('\n');
        cstr = cstr.mid(index + 1);
    }
    emitDatLineList(cstr);
}
