/***************************************************************************
*   Copyright (C) 2004 by Kita Developers                                 *
*   ikemo@users.sourceforge.jp                                            *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
***************************************************************************/

#include "datinfo.h"

#include <QtCore/QDir>
#include <QtCore/QRegExp>
#include <QtCore/QStringList>

#include <klocale.h>

#include "abone.h"
#include "access.h"
#include "account.h"
#include "cache.h"
#include "datmanager.h"
#include "globalconfig.h"
#include "kita_misc.h"
#include "kita-utf8.h"
#include "parser.h"
#include "thread.h"

using namespace Kita;

static const int RESDAT_DEFAULTSIZE = 10;
static const int RESDAT_DELTA = 1000;


/*------------------------------------------------------*/
/* DatInfo stores & handles all information about *.dat */

DatInfo::DatInfo(const KUrl& url) : m_threadIndex(m_datUrl),
    m_access (0), m_access2(0)
{
    QString refstr;
    m_datUrl = getDatUrl(url, refstr);
    m_threadIndex = ThreadIndex(m_datUrl);
    // get the pointer of Thread class
    m_thread = Thread::getByUrlNew(m_datUrl);
    if (m_thread == 0) {

        // create Thread
        m_thread = Thread::getByUrl(m_datUrl);
        if (m_thread == 0) return ;

        // read idx file
        m_threadIndex.loadIndex(m_thread);
    }

    m_thread = Thread::getByUrl(m_datUrl);

    // japanese strings
    m_spacestr = QString::fromUtf8(KITAUTF8_ZENSPACE);
    m_framestr1 = QString::fromUtf8(KITAUTF8_FRAME1); // |
    m_framestr2 = QString::fromUtf8(KITAUTF8_FRAME2); // |-
    m_framestr3 = QString::fromUtf8(KITAUTF8_FRAME3); // L

    // make directory
    Cache cache(m_datUrl);
    QString cacheDir = cache.getDirPath();
    if (!QDir::root().mkpath(cacheDir)) return ;

    initPrivate(
        true // load cache
   );
}

DatInfo::~DatInfo()
{
    initPrivate(false);
}


/* Init  */
/* Usually, don't call this. */ /* public */
void DatInfo::init()
{
    return initPrivate(true);
}

/* Init. If loadCache = true, load data from cache. */ /* private */
void DatInfo::initPrivate(bool loadCache)
{
    // stop & delete dat loader
    deleteAccessJob();

    // init variables
    m_broken = false;
    m_nowLoading = false;
    m_lastLine.clear();

    // clear ResDatVec
    m_resDatVec.clear();
    increaseResDatVec(RESDAT_DEFAULTSIZE);

    // reset Abone
    resetAbonePrivate();

    // create dat loader
    m_access = new Access(m_datUrl);

    connect(m_access, SIGNAL(receiveData(const QStringList&)),
             SLOT(slotReceiveData(const QStringList&)));
    connect(m_access, SIGNAL(finishLoad()), SLOT(slotFinishLoad()));

    if (!loadCache) return ;

    // reset ReadNum before loading cache.
    // ReadNum & subject are updated by Access::getcache()
    m_thread->setReadNum(0);

    // get dat from cahce
    // slotReceiveData() is called from Access::getcache()
    m_access->getcache();

    // save up-to-date thread information
    m_threadIndex.saveIndex(m_thread);
}


/* private */
void DatInfo::resetResDat(RESDAT& resdat) const
{
    resdat.num = 0;
    resdat.parsed = false;
    resdat.broken = false;
    resdat.anclist.clear();
    resdat.checkAbone = false;
    resdat.abone = false;
    resdat.isResponsed = false;
}


/* private */
void DatInfo::increaseResDatVec(int delta)
{
    int size = m_resDatVec.size();
    RESDAT resdat;
    resetResDat(resdat);
    m_resDatVec.resize(size + delta);
    for (int i = size; i < size + delta; i++)
        m_resDatVec[i] = resdat;
}


/* delete dat loader */ /* private */
void DatInfo::deleteAccessJob()
{
    if (m_access) {
        m_access->killJob();
        delete m_access;
        m_access = 0;
    }
    if (m_access2) {
        m_access2->killJob();
        delete m_access2;
        m_access2 = 0;
    }
}

/* public */
const KUrl& DatInfo::url() const
{
    return m_datUrl;
}



/*--------------------------------------*/
/* cache handling functions             */

/* Update cache  */

/* When Access received new data,
   slotReceiveData is called.           */

/* When Access fineshed loading,
   slotFinishLoad is called, and
   DatInfo emits the finishLoad signal to the parent object  */ /* public */
bool DatInfo::updateCache(const QObject* parent)
{
    if (m_access == 0)
        return false;
    if (m_nowLoading)
        return false;

    m_nowLoading = true;

    connect(this, SIGNAL(receiveData()),
             parent, SLOT(slotReceiveData()));

    connect(this, SIGNAL(finishLoad()),
             parent, SLOT(slotFinishLoad()));

    m_access->getupdate(m_thread->readNum());

    return true;
}


/* slot called when Access
   received new data              */      /* private  slot */
void DatInfo::slotReceiveData(const QStringList& lineList)
{
    int rescode = m_access->responseCode();
    if (m_access2) {
        rescode = m_access2->responseCode();
    }

    if (rescode != 200 && rescode != 206) return ;

    // copy lines to buffer
    int count = lineList.count();
    for (int i = 0; i < count ; ++i)
        copyOneLineToResDat(lineList[ i ]);

    emit receiveData();
}


/* copy one line to resdat.
   See also DatInfo::slotReceiveData()   */ /* private */
bool DatInfo::copyOneLineToResDat(const QString& line)
{
    if (line.isEmpty()) return false;

    // update ReadNum
    const int num = m_thread->readNum() + 1;
    m_thread->setReadNum(num);

    // If resdat vector is short, then resize the vector.
    while ((int) m_resDatVec.size() <= num)
        increaseResDatVec(RESDAT_DELTA);

    // reset ResDat
    RESDAT& resdat = m_resDatVec[ num ];
    resetResDat(resdat);

    resdat.num = num;
    resdat.linestr = line;

    // get subject
    if (num == 1) parseDat(num);

    // search all responses which are responsed by this line.
    if (GlobalConfig::checkResponsed()) {

        if (parseDat(num) && !checkAbonePrivate(num)) { // parse line here to get AncList

            const int maxRange = 10;

            AncList& anclist = m_resDatVec[ num ].anclist;
            for (AncList::iterator it = anclist.begin(); it != anclist.end(); ++it) {

                int fromNum = (*it).from;
                int toNum = qMin(num - 1, (*it).to);
                if (toNum - fromNum + 1 > maxRange) continue;

                for (int i = fromNum; i <= toNum; ++i) {

                    if (!checkAbonePrivate(i)) m_resDatVec[ i ].isResponsed = true;
                }
            }
        }
    }

    return true;
}


/* slot called when Access
   finished loading new dat */      /* private  slot */
void DatInfo::slotFinishLoad()
{
    // save thread information
    m_threadIndex.saveIndex(m_thread);

    // re-try by offlaw.cgi
    DatManager datManager(m_datUrl);
    if (m_thread->readNum() == 0 && m_access2 == 0
            && datManager.is2chThread()) {
        if (Account::isLogged()) {
            initPrivate(true);
            m_access2 = new OfflawAccess(m_datUrl);
            connect(m_access2, SIGNAL(receiveData(const QStringList&)),
                     SLOT(slotReceiveData(const QStringList&)));
            connect(m_access2, SIGNAL(finishLoad()), SLOT(slotFinishLoad()));
            m_access2->get();
            return ;
        }
    }
    // finish loading session & emit signal to the parent object
    m_nowLoading = false;
    emit finishLoad();

    // disconnect signals
    disconnect(SIGNAL(receiveData()));
    disconnect(SIGNAL(finishLoad()));
}


/* public */
int DatInfo::getResponseCode() const
{
    return (m_access == 0) ? 0 : m_access->responseCode();
}


/* public */
int DatInfo::getServerTime() const
{
    return (m_access == 0) ? 0 : m_access->serverTime();
}


/* public */
bool DatInfo::deleteCache()
{
    if (m_nowLoading)
        return false;

    initPrivate(false);

    return true;
}


/* public */
bool DatInfo::isLoadingNow() const
{
    return m_nowLoading;
}



/* public */
void DatInfo::stopLoading() const
{

    // Don't lock the mutex here !!!
    // It will cause deadlock , because
    // Access::stopJob() calls HTMLPart::slotFinishLoad() back,
    // then HTMLPart::slotFinishLoad() calls another functions in DatInfo.
    if (m_access == 0)
        return;
    if (!m_nowLoading)
        return;

    m_access->stopJob();
}

/*------------------------------------------------------*/
/* get subject, linedata,  id, body, name, HTML, etc.   */

/* They are public */

QString DatInfo::getDat(int num)
{
    return (!parseDat(num)) ? QString() : m_resDatVec[ num ].linestr;
}

QString DatInfo::getId(int num)
{
    return (!parseDat(num)) ? QString() : m_resDatVec[ num ].id;
}

/* plain strings of name */
QString DatInfo::getPlainName(int num)
{
    return (!parseDat(num)) ? QString() : m_resDatVec[ num ].name;
}


/* plain strings of title */
QString DatInfo::getPlainTitle(int num)
{
    if (!parseDat(num)) return QString();

    QString titleHtml;
    createTitleHtml(m_resDatVec[ num ], titleHtml);

    QString retStr;
    Parser::datToText(titleHtml, retStr);

    return retStr;
}


/* plain strings of body  */
QString DatInfo::getPlainBody(int num)
{
    if (!parseDat(num)) return QString();

    QString retStr;
    Parser::datToText(m_resDatVec[ num ].bodyHTML, retStr);

    return retStr;
}


/*-----------------------------------------*/
/* HTML data                               */


/* get HTML strings of title & body.
 
   return values are defined in datinfo.h.  */ /* public */

int DatInfo::getHTML(int num, bool checkAbone, QString& titleHTML,
        QString& bodyHTML)
{
    return getHTMLPrivate(num, checkAbone, titleHTML, bodyHTML);
}

/**
 * @param[in] num
 * @param[in] checkAbone
 *
 * @param[out] titleHTML
 * @param[out] bodyHTML
 *
 * @retval HTML_NOTPARSED The dat is not parsed.
 * @retval HTML_ABONE The res dat is marked as abone.
 * @retval HTML_BROKEN The res dat is marked as broken.
 * @retval HTML_NORMAL The res dat is normal.
 *
 */
int DatInfo::getHTMLPrivate(int num, bool checkAbone, QString& titleHtml,
        QString& bodyHtml)
{
    if (!parseDat(num)) return HTML_NOTPARSED;

    bool abone = checkAbone & checkAbonePrivate(num);
    RESDAT& resdat = m_resDatVec[ num ];

    if (abone) {
        titleHtml = QString::number(num) + ' ' + i18n("Abone");
        bodyHtml = "<a href=\"#abone" + QString::number(num) + "\">";
        bodyHtml += i18n("Abone") + "</a>";

        return HTML_ABONE;
    } else if (resdat.broken) {
        titleHtml = QString::number(num) + ' ' + i18n("Broken");
        bodyHtml = i18n("Broken");

        return HTML_BROKEN;
    } else {
        createTitleHtml(resdat, titleHtml);
        bodyHtml = resdat.bodyHTML;
        
        return HTML_NORMAL;
    }
}

/* get HTML strings from startnum to endnum.
 
   return value is HTML strings               */  /* public */
QString DatInfo::getHTMLString(int startnum, int endnum, bool checkAbone)
{
    QString retHTML;

    for (int num = startnum; num <= endnum; num++) {

        QString html;
        getHtmlOfOneRes(num, checkAbone, html);
        retHTML += html;
    }

    return retHTML;
}


/* return HTML strings that have ID = strid. */ /* public */
QString DatInfo::getHtmlById(const QString& strid, int &count)
{
    QString retHTML;
    count = 0;

    for (int i = 1; i <= m_thread->readNum(); i++) {

        if (!parseDat(i)) continue;

        if (m_resDatVec[ i ].id == strid) {
            count ++;

            QString html;
            getHtmlOfOneRes(i, true, html);
            retHTML += html;
        }
    }

    return retHTML;
}

/**
 *
 * convert res dat to html.
 *
 * @param[in] num
 * @param[in] checkAbone
 *
 * @param[out] html
 *
 */
void DatInfo::getHtmlOfOneRes(int num, bool checkAbone, QString& html)
{
    html.clear();
    QString titleHTML, bodyHTML;
    if (getHTMLPrivate(num, checkAbone, titleHTML, bodyHTML) == HTML_NOTPARSED)
        return;

    if (m_resDatVec[ num ].isResponsed)
        titleHTML.replace(
                "<a href=\"#write", "<a class=\"coloredLink\" href=\"#write");
    html += "<div class=\"pop_res_title\">" + titleHTML + "</div>";
    html += "<div class=\"pop_res_body\">" + bodyHTML + "</div>";
}



/*-------------------------------*/
/* Get HTML document of res tree.*/
/* For example, when rootnum = 1,
 
>>1 
|-->>4
|  |--->>10
|
|-->>20, and return count = 3.  */

/* Note that this function checks Abone internally. */ /* public */
QString DatInfo::getTreeByRes(int rootnum, int& count)
{
    return getTreeByResPrivate(rootnum, false, count);
}

/*---------------------------------------*/
/* Get HTML document of reverse res tree.*/
/* For example, when rootnum = 10,
 
>>10 
|-->>5
|  |--->>2
|
|-->>6, and returns count = 3.  */

/* Note that this function checks Abone internally. */ /* public */
QString DatInfo::getTreeByResReverse(int rootnum, int& count)
{
    return getTreeByResPrivate(rootnum, true, count);
}


/* private */
QString DatInfo::getTreeByResPrivate(int rootnum,
    bool reverse /* reverse search */, int& count)
{

    QString tmp = QString::number(rootnum);
    QString retstr = "<a href=\"#" + tmp + "\">&gt;&gt;" + tmp + "</a><br>";

    retstr += getTreeByResCore(rootnum, reverse, count, "");

    return retstr;
}

/* private */
QString DatInfo::getTreeByResCore(int rootnum,
        bool reverse /* reverse search */, int& count, const QString& prestr)
{
    if (!parseDat(rootnum))
        return QString();
    if (checkAbonePrivate(rootnum))
        return QString();

    QString retstr;
    count = 0;
    QStringList strlists;

    if (!reverse) {

        // collect responses that have anchor to rootnum
        for (int i = rootnum + 1; i <= m_thread->readNum(); i++) {
            if (checkAbonePrivate(i)) continue;
            if (checkRes(i, rootnum)) {
                count ++;
                strlists += QString::number(i);
            }
        }

    } else { // collect responses for which rootnum has anchors

        AncList& anclist = m_resDatVec[ rootnum ].anclist;
        for (AncList::iterator it = anclist.begin(); it != anclist.end(); ++it) {
            for (int i = (*it).from; i <= qMin(rootnum - 1, (*it).to) ; i++) {
                if (checkAbonePrivate(i)) continue;
                count ++;
                strlists += QString::number(i);
            }
        }
    }

    // make HTML document
    if (count) {

        for (QStringList::iterator it = strlists.begin(); it != strlists.end(); ++it) {
            QString tmpstr;
            if ((*it) == strlists.last()) tmpstr = m_framestr3;  // 'L'
            else tmpstr = m_framestr2;  // '|-'

            retstr += prestr + tmpstr + "<a href=\"#" + (*it) + "\">&gt;&gt;" + (*it) + "</a><br>";

            // call myself recursively
            int tmpnum;
            tmpstr = prestr;
            if ((*it) == strlists.last()) tmpstr += m_spacestr + m_spacestr + m_spacestr; // "   "
            else tmpstr += m_framestr1 + m_spacestr; // "| "
            retstr += getTreeByResCore((*it).toInt(), reverse, tmpnum, tmpstr);
            count += tmpnum;
        }
    }

    return retstr;
}



/*----------------------------------------------*/
/* Check if No.num has anchors to No.target     */
/* For exsample, if target = 4, and No.num have
   an anchor >>4, or >>2-6, etc.,
   then return true.                            */  /* private */
bool DatInfo::checkRes(int num, int target)
{
    const int range = 20;
    if (!parseDat(num))
        return false;

    AncList& anclist = m_resDatVec[ num ].anclist;

    for (AncList::iterator it = anclist.begin(); it != anclist.end(); ++it) {
        if ((*it).to - (*it).from > range)
            continue;
        if (target >= (*it).from && target <= (*it).to)
            return true;
    }

    return false;

}


/*-----------------------*/
/* several information */

int DatInfo::getResNum() const
{
    return m_thread->resNum();
}

/* public */
int DatInfo::getReadNum() const
{
    return m_thread->readNum();
}

/* public */
int DatInfo::getViewPos() const
{
    return m_thread->viewPos();
}


/* return number of responses that have ID = strid. */
/* Note that this function checks Abone internally. */ /* public */
int DatInfo::getNumById(const QString& strid)
{
    int count = 0;

    for (int i = 1; i <= m_thread->readNum(); i++) {

        if (!parseDat(i))
            continue;
        if (checkAbonePrivate(i))
            continue;

        if (m_resDatVec[ i ].id == strid)
            count++;
    }

    return count;
}


/* public */
int DatInfo::getDatSize() const
{
    return (m_access == 0) ? 0 : m_access->dataSize();
}


/* public */
bool DatInfo::isResponsed(int num) const
{
    return m_resDatVec[ num ].isResponsed;
}


/* public */
bool DatInfo::isResValid(int num)
{
    return parseDat(num);
}

/* public */
bool DatInfo::isBroken() const
{
    if (m_broken)
        return m_broken;

    if (m_access == 0)
        return false;

    int rescode = m_access->responseCode();
    bool invalid = m_access->invalidDataReceived();

    // see also Access::slotReceiveThreadData()
    if (invalid && (rescode == 200 || rescode == 206)) return true;

    // maybe "Dat Ochi"
    return false;
}

/* public */
bool DatInfo::isResBroken(int num)
{
    return (!parseDat(num)) ? false : m_resDatVec[ num ].broken;
}

/* ID = strid ? */ /* public */
bool DatInfo::checkId(const QString& strid, int num)
{
    if (!parseDat(num))
        return false;

    if (m_resDatVec[ num ].id == strid)
        return true;

    return false;
}


/* Are keywords included? */ /* public */
bool DatInfo::checkWord(const QStringList& stlist,   /* list of keywords */
        int num, bool checkOr /* AND or OR search */)
{
    if (!parseDat(num))
        return false;

    QString str_text = m_resDatVec[ num ].bodyHTML;

    for (QStringList::const_iterator it = stlist.begin(); it != stlist.end(); ++it) {

        QRegExp regexp((*it));
        regexp.setCaseSensitivity(Qt::CaseInsensitive);

        if (checkOr) { // OR
            if (str_text.indexOf(regexp, 0) != -1) {
                return true;
            }
        } else { // AND
            if (str_text.indexOf(regexp, 0) == -1) return false;
        }
    }

    if (checkOr) return false;

    return true;
}



/*--------------------------------*/
/* abone functions                */


/*-----------------------*/
/* reset abone.          */

/* call this when config
   of abone changed.     */  /* public */

void DatInfo::resetAbone()
{
    return resetAbonePrivate();
}

/* private */
void DatInfo::resetAbonePrivate()
{
    for (int i = 1; i < (int) m_resDatVec.size(); i++)
        m_resDatVec[ i ].checkAbone = false;

    m_aboneByID = (!AboneConfig::aboneIDList().isEmpty());
    m_aboneByName = (!AboneConfig::aboneNameList().isEmpty());
    m_aboneByBody = (!AboneConfig::aboneWordList().isEmpty());
    m_aboneChain = (m_aboneByID | m_aboneByName | m_aboneByBody)
        & GlobalConfig::aboneChain() ;
}


/*--------------*/
/* check abone  */ /* public */

bool DatInfo::checkAbone(int num)
{
    return checkAbonePrivate(num);
}


/* private */
bool DatInfo::checkAbonePrivate(int num)
{
    if (!parseDat(num))
        return false;

    if (m_resDatVec[num].checkAbone)
        return m_resDatVec[num].abone;

    m_resDatVec[num].checkAbone = true;
    bool checktmp = false;

    if (m_aboneByID)
        checktmp = checkAboneCore(m_resDatVec[num].id,
                AboneConfig::aboneIDList());

    if (!checktmp && m_aboneByName)
        checktmp = checkAboneCore(m_resDatVec[num].name,
                AboneConfig::aboneNameList());

    if (!checktmp && m_aboneByBody)
        checktmp = checkAboneCore(m_resDatVec[num].bodyHTML,
                AboneConfig::aboneWordList());

    if (!checktmp && m_aboneChain) {
        AncList & anclist = m_resDatVec[num].anclist;

        for (AncList::iterator it = anclist.begin();
                it != anclist.end() && !checktmp ; ++it) {

            int refNum = (*it).from;
            int refNum2 = (*it).to;

            // I don't want to enter loop...
            if (refNum >= num) continue;
            if (refNum2 >= num) refNum2 = num - 1;

            for (int i = refNum; i <= refNum2; i++) {
                if (checkAbonePrivate(i)) {
                    checktmp = true;
                    break;
                }
            }
        }
    }

    m_resDatVec[num].abone = checktmp;

    return m_resDatVec[num].abone;
}

/* private */
bool DatInfo::checkAboneCore(const QString& str, const QStringList& strlist)
const
{
    if (strlist.count()) {

        int i;
        for (QStringList::const_iterator it = strlist.begin();
                it != strlist.end(); ++it) {
            i = str.indexOf((*it));
            if (i != -1) {
                return true;
            }
        }
    }

    return false;
}



/* parsing function for ResDat */
/* This function parses the raw data by parseResDat() */ /* private */
bool DatInfo::parseDat(int num)
{
    if (num <= 0 || m_thread->readNum() < num)
        return false;
    if (m_resDatVec[ num ].parsed)
        return true;

    //   qDebug("parseDat %d",num);

    QString subject;
    Parser::parseResDat(m_resDatVec[ num ], subject);
    if (num == 1 && !subject.isEmpty())
        m_thread->setThreadName(subject);
    if (m_resDatVec[ num ].broken)
        m_broken = true;

    return true;
}

bool DatInfo::isOpened() const
{
    return m_isOpened;
}

void DatInfo::setOpened(bool isOpened)
{
    m_isOpened = isOpened;
}

/* create HTML of title.
  
  struct RESDAT resdat should be parsed by parseResDat before  calling this function.
 
  output: titleHtml
  
*/
void DatInfo::createTitleHtml(RESDAT& resdat, QString& titleHtml)
{
    titleHtml.clear();
    if (!resdat.parsed) return ;

    bool showMailAddress = GlobalConfig::showMailAddress();
    bool useTableTag = GlobalConfig::useStyleSheet();

    if (useTableTag) titleHtml += "<table class=\"res_title\"><tr>";

    // res number
    if (useTableTag) titleHtml += "<td class=\"res_title_number\">";
    titleHtml += "<a href=\"#write" + QString::number(resdat.num) + "\">";
    titleHtml += QString::number(resdat.num);
    titleHtml += "</a> ";

    // name & mail address
    if (useTableTag) titleHtml += "<td class=\"res_title_name\">";
    titleHtml += "<b>" + QString::fromUtf8(KITAUTF8_NAME);

    // show name with mail address
    if (showMailAddress) {

        titleHtml += resdat.nameHTML;
        if (!resdat.address.isEmpty()) titleHtml += " [" + resdat.address + ']';

    } else { // don't show mail address

        if (resdat.address.isEmpty()) {

            titleHtml += "<span class=\"name_noaddr\">";
            titleHtml += resdat.name;
            titleHtml += "</span>";

        } else {

            titleHtml += "<a href=\"mailto:" + resdat.address + "\"";
            titleHtml += " title=\"" + resdat.address + "\">";
            titleHtml += resdat.name;
            titleHtml += "</a>";
        }
    }

    titleHtml += "</b> ";

    // date
    if (useTableTag) titleHtml += "<td class=\"res_title_date\">";
    titleHtml += QString::fromUtf8(KITAUTF8_COLON) + resdat.date;
    if (useTableTag) titleHtml += "</td>";

    // ID
    if (!resdat.id.isEmpty()) {

        if (useTableTag) titleHtml += "<td class=\"res_title_id\">";
        if (resdat.id.count("???") >= 1) titleHtml += " ID:" + resdat.id;
        else titleHtml += " <a href=\"#idpop" + resdat.id + "\">ID</a>" + ":" + resdat.id;
        if (useTableTag) titleHtml += "</td>";
    }

    // BE
    if (!resdat.be.isEmpty()) {

        if (useTableTag) titleHtml += "<td class=\"res_title_be\">";
        titleHtml += " <a href=\"#bepop" + resdat.be + "\">?" + resdat.bepointmark + "</a>";
        if (useTableTag) titleHtml += "</td>";
    }

    // host
    if (!resdat.host.isEmpty()) {

        if (useTableTag) titleHtml += "<td class=\"res_title_host\">";
        titleHtml += " HOST:" + resdat.host;
        if (useTableTag) titleHtml += "</td>";
    }

    if (useTableTag) titleHtml += "</tr></table>";
}
