/***************************************************************************
*   Copyright (C) 2004 by Kita Developers                                 *
*   ikemo@users.sourceforge.jp                                            *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
***************************************************************************/

/* DatManager manages all information about thread */
#include "datmanager.h"

#include <QtCore/QFile>
#include <QtCore/QObject>
#include <QtCore/QRegExp>
#include <QtCore/QStringList>

#include "boarddatabase.h"
#include "cache.h"
#include "datinfo.h"
#include "kita_misc.h"
#include "thread.h"
#include "threadindex.h"
#include "threadinfo.h"

using namespace Kita;

static const int DMANAGER_MAXQUEUE = 16;

DatInfoList DatManager::m_datInfoList;


/*-----------------------------------------------------------------------*/
DatManager::DatManager(const KUrl& url) : m_url(url)
{
    m_datUrl = getDatUrl(url);
    m_searchDatInfo = searchDatInfo();
    m_datInfo = getDatInfo();
}


/* create DatInfo explicitly.                    */
/* Usually, DatInfo is NOT created
   if cache does not exist(i.e. ReadNum == 0). */ /* public */
bool DatManager::createDatInfo() const
{
    return getDatInfo(false /* don't check the existence of cache */) != 0;
}


/* get pointer of DatInfo */

/*    !!!  NOTICE  !!!   */
/* It is very dangerous to access to DatInfo directly. */
/* Usually, access to it through DatManager.           */ /* public */
DatInfo * DatManager::getDatInfoPointer() const
{
    return m_datInfo;
}


/*------------------------------------------------------------------------*/


/* This function searches instance of DatInfo in m_datInfoList.
 
   If cache exists, create DatInfo and return its pointer.
 
   If checkCached == true and cache does not exist, return 0.
 
   If checkCached == false and cache does not exist,
   create DatInfo and return its pointer anyway.
 
   see also DatManager::searchDatInfo() and DatManager::createDatInfo() */ /* private */

DatInfo* DatManager::getDatInfo(bool checkCached) const
{
    // search
    // create and record instance
    return (m_searchDatInfo != 0)
        ? m_searchDatInfo : recordDatInfo(checkCached);
}


/* This function just searches instance of DatInfo specified by datURL
   without creating instance.  */ /* private */
DatInfo* DatManager::searchDatInfo() const
{
    if (m_datUrl.isEmpty())
        return 0; // This url is not recorded in BoardManager.
    if (m_datInfoList.isEmpty())
        return 0;

    int i = 0;
    DatInfoList::Iterator it;
    DatInfo* datInfo;

    for (it = m_datInfoList.begin(); it != m_datInfoList.end(); ++it, i++) {

        datInfo = (*it);

        if (m_datUrl == datInfo->url()) {

            // LRU
            if (i) {
                m_datInfoList.erase(it);
                m_datInfoList.prepend(datInfo);
            }

            return datInfo;
        }
    }

    return 0;
}


/* create and record the instance of DatInfo and delete old instances.
   Note that DatInfo::DatInfo() opens cached data and reads it. */
/* private */
DatInfo* DatManager::recordDatInfo(bool checkCached) const
{
    if (m_datUrl.isEmpty())
        return 0; // This url is not recorded in BoardManager.

    // create DatInfo & read cached data
    DatInfo* datInfo = new DatInfo(m_datUrl);

    // Does cache exist?
    // If cache does not exist, delete DatInfo here.
    if (checkCached && datInfo->getReadNum() == 0) {
        delete datInfo;
        return 0;
    }

    m_datInfoList.prepend(datInfo);

    // delete the all old instances (LRU algorithm)
    if (m_datInfoList.count() > DMANAGER_MAXQUEUE) {
        for (int i = DMANAGER_MAXQUEUE; i < m_datInfoList.count(); i++) {
            DatInfo* deleteInfo = m_datInfoList.at(i);
            if (deleteInfo == 0)
                continue;
            m_datInfoList.removeAt(i);
            i--;
            delete datInfo;
        }
    }

    return datInfo;
}


/* public */
void DatManager::deleteAllDatInfo()
{
    while (!m_datInfoList.isEmpty())
        delete m_datInfoList.takeFirst();
}



/*-------------------------------------------------------------*/



/* update cache */   /* public */
bool DatManager::updateCache(const QObject* parent) const
{
    return (m_datInfo == 0) ? false : m_datInfo->updateCache(parent);
}


/* public */
int DatManager::getResponseCode() const
{
    return (m_datInfo == 0) ? 0 : m_datInfo->getResponseCode();
}

/* public */
int DatManager::getServerTime() const
{
    return (m_datInfo == 0) ? 0 : m_datInfo->getServerTime();
}


/* public */
bool DatManager::deleteCache() const
{
    Thread* thread = Thread::getByUrlNew(m_datUrl);
    if (thread == 0)
        return false;
    if (thread->readNum() == 0)
        return false;

    // init DatInfo
    if (m_searchDatInfo && !m_searchDatInfo->deleteCache()) {
        return false;
    }

    // reset readNum & veiwPos
    thread->setReadNum(0);
    thread->setViewPos(0);

    // delete cache
    Cache cache(m_datUrl);
    QString cachePath = cache.getPath();
    QString indexPath = cache.getIndexPath();
    QFile::remove(indexPath);
    QFile::remove(cachePath);

    // delete log from "cache"
    ThreadInfo::removeThreadInfo(m_datUrl.prettyUrl());
    return true;
}


/* public */
bool DatManager::isLoadingNow() const
{
    return (m_searchDatInfo == 0) ? false : m_searchDatInfo->isLoadingNow();
}


/* public */
void DatManager::stopLoading() const
{
    if (m_searchDatInfo == 0)
        return;
    m_searchDatInfo->stopLoading();
}

/*--------------------------------------*/
/* string data */


/* public */
QString DatManager::getDat(int num) const
{
    return (m_datInfo == 0) ? QString() : m_datInfo->getDat(num);
}



/* public */
QString DatManager::getId(int num) const
{
    return (m_datInfo == 0) ? QString() : m_datInfo->getId(num);
}


/* public */
QString DatManager::getPlainName(int num) const
{
    return (m_datInfo == 0) ? QString() : m_datInfo->getPlainName(num);
}


/* public */
QString DatManager::getPlainBody(int num) const
{
    return (m_datInfo == 0) ? QString() : m_datInfo->getPlainBody(num);
}


/* public */
QString DatManager::getPlainTitle(int num) const
{
    return (m_datInfo == 0) ? QString() : m_datInfo->getPlainTitle(num);
}


/* get name (i.e. subject) of thread from URL of dat file. */ /* public */
QString DatManager::threadName() const
{
    Thread* thread = Thread::getByUrlNew(m_datUrl);
    return (thread != 0) ? thread->threadName() : QString();
}


/* public */
QString DatManager::threadId() const
{
    return m_datUrl.fileName().section('.', 0, 0);
}

/*---------------------------------------*/
/* HTML data */

/* public */
QString DatManager::getHtml(int startnum, int endnum, bool checkAbone) const
{
    return (m_datInfo == 0)
        ? QString() : m_datInfo->getHTMLString(startnum, endnum, checkAbone);
}



/* public */
QString DatManager::getHtmlById(const QString& strid, int &count) const
{
    return (m_datInfo == 0) ? QString() : m_datInfo->getHtmlById(strid, count);
}



/* Get HTML document of res tree.*/ /* public */
QString DatManager::getTreeByRes(int rootnum, int &count) const
{
    return (m_datInfo == 0)
        ? QString() : m_datInfo->getTreeByRes(rootnum, count);
}

/* Get HTML document of reverse res tree.*/ /* public */
QString DatManager::getTreeByResReverse(int rootnum, int &count) const
{
    return (m_datInfo == 0)
        ? QString() : m_datInfo->getTreeByResReverse(rootnum, count);
}


/* public */
int DatManager::getResNum() const
{
    Thread* thread = Thread::getByUrlNew(m_datUrl);
    return (thread != 0) ? thread->resNum() : 0;
}


/* public */
int DatManager::getReadNum() const
{
    Thread* thread = Thread::getByUrlNew(m_datUrl);
    return (thread != 0) ? thread->readNum() : 0;
}


/* public */
int DatManager::getViewPos() const
{
    Thread* thread = Thread::getByUrlNew(m_datUrl);
    return (thread != 0) ? thread->viewPos() : 0;
}


/* public */
void DatManager::setViewPos(int num) const
{
    Thread* thread = Thread::getByUrlNew(m_datUrl);
    if (thread != 0)
        thread->setViewPos(num);

    // save idx
    ThreadIndex threadIndex(m_url);
    threadIndex.setViewPos(num);

    // save "cache"
    ThreadInfo::setReadNum(m_datUrl.prettyUrl(), num);
}


/* public */
int DatManager::getDatSize() const
{
    return (m_datInfo == 0) ? 0 : m_datInfo->getDatSize();
}

/* get number of responses which have same ID. */ /* public */
int DatManager::getNumById(const QString& strid) const
{
    return (m_datInfo == 0) ? 0 : m_datInfo->getNumById(strid);
}


/* public */
bool DatManager::isThreadRecorded() const
{
    return !m_datUrl.isEmpty();
}


/* public */
bool DatManager::is2chThread() const
{
    BoardDatabase db(m_url);
    if (db.type() != Board_2ch)
        return false;
    if (m_datUrl.isEmpty())
        return false;

    QRegExp url_2ch(".*\\.2ch\\.net");
    QRegExp url_bbspink(".*\\.bbspink\\.com");

    if (url_2ch.indexIn(m_url.host()) != -1
            || url_bbspink.indexIn(m_url.host()) != -1)
        return true;

    return false;
}


/* public */
bool DatManager::isResValid(int num) const
{
    return (m_datInfo == 0) ? false : m_datInfo->isResValid(num);
}


/* public */
bool DatManager::isBroken() const
{
    return (m_datInfo == 0) ? false : m_datInfo->isBroken();
}

/* public */
bool DatManager::isResBroken(int num) const
{
    return (m_datInfo == 0) ? false : m_datInfo->isResBroken(num);
}



/* check if ID == strid  */ /* public */
bool DatManager::checkId(const QString& strid, int num) const
{
    return (m_datInfo == 0) ? false : m_datInfo->checkId(strid, num);
}


/* check if keywords are included */ /* public */
bool DatManager::checkWord(const QStringList& strlist, int num,
        bool checkOr /* AND or OR search */) const
{
    return (m_datInfo == 0)
        ? false : m_datInfo->checkWord(strlist, num, checkOr);
}


/* public */
bool DatManager::isMarked(int num) const
{
    Thread* thread = Thread::getByUrlNew(m_datUrl);
    return (thread == 0) ? false : thread->isMarked(num);
}


/* public */
void DatManager::setMark(int num, bool mark) const
{
    Thread* thread = Thread::getByUrlNew(m_datUrl);
    if (thread == 0)
        return;

    if (thread->setMark(num, mark)) {
        ThreadIndex threadIndex(m_url);
        threadIndex.setMarkList(thread->markList());
    }
}


/* public */
bool DatManager::checkAbone(int num) const
{
    return (m_datInfo == 0) ? false : m_datInfo->checkAbone(num);
}


/* public */
void DatManager::resetAbone() const
{
    if (m_datInfo == 0)
        return;
    m_datInfo->resetAbone();
}


/* check if the thread is shown on the main thread tab. */ /* public */
bool DatManager::isMainThreadOpened() const
{
    return (m_datInfo == 0) ? false : m_datInfo->isOpened();
}

void DatManager::setMainThreadOpened(bool isOpened) const
{
    if (m_datInfo == 0)
        return;
    m_datInfo->setOpened(isOpened);
}


/*--------------------------*/
/* obsolete */

/* public */
QString DatManager::threadUrl() const
{
    return getThreadUrl(m_url);
}
