/***************************************************************************
*   Copyright (C) 2004 by Kita Developers                                 *
*   ikemo@users.sourceforge.jp                                            *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
***************************************************************************/

#include "threadindex.h"

#include <QtCore/QFile>

#include <kconfig.h>
#include <KConfigGroup>

#include "cache.h"
#include "datmanager.h"
#include "kita_misc.h"
#include "thread.h"
#include "threadinfo.h"

using namespace Kita;

ThreadIndex::ThreadIndex(const KUrl& url) : m_url(url)
{
    Cache cache(m_url);
    m_indexPath = cache.getIndexPath();
}

QString ThreadIndex::getSubject() const
{
    KConfig config(m_indexPath);
    return getSubjectPrivate(config);
}

void ThreadIndex::setSubject(const QString& str)
{
    KConfig config(m_indexPath);
    setSubjectPrivate(config, str);
}

int ThreadIndex::getResNum() const
{
    KConfig config(m_indexPath);
    return getResNumPrivate(config);
}

void ThreadIndex::setResNum(int resNum)
{
    KConfig config(m_indexPath);
    setResNumPrivate(config, resNum);
}

int ThreadIndex::getReadNum() const
{
    KConfig config(m_indexPath);
    return getReadNumPrivate(config, true);
}

void ThreadIndex::setReadNum(int readNum)
{
    KConfig config(m_indexPath);
    setReadNumPrivate(config, readNum);
}

int ThreadIndex::getViewPos() const
{
    KConfig config(m_indexPath);
    return getViewPosPrivate(config);
}

void ThreadIndex::setViewPos(int viewPos)
{
    KConfig config(m_indexPath);
    setViewPosPrivate(config, viewPos);
}

void ThreadIndex::setMarkList(const QList<int>& markList)
{
    KConfig config(m_indexPath);
    setMarkListPrivate(config, markList);
}


/*------------------------------------------------------------*/


/* load thread information */  /* public */
void ThreadIndex::loadIndex(Thread* thread, bool checkCached) const
{
    KConfig config(m_indexPath);
    // load read number
    int readNum = getReadNumPrivate(config, checkCached);
    if (readNum == 0) return ;  // cache does not exist.
    thread->setReadNum(readNum);

    // load thread name
    QString subject = getSubjectPrivate(config);
    if (subject.isEmpty() && !thread->threadName().isEmpty()) {
        subject = thread->threadName();
        KConfigGroup group = config.group("");
        group.writeEntry("Subject", subject);
    }
    if (subject.isEmpty()) thread->setThreadName("?");
    else thread->setThreadName(subject);

    // load res number
    thread->setResNum(getResNumPrivate(config));

    // load view pos
    thread->setViewPos(getViewPosPrivate(config));
    if (thread->viewPos() > thread->readNum())
        thread->setReadNum(thread->viewPos());

    // load mark
    thread->setMarkList(getMarkListPrivate(config));
}



/* save thread information */  /* public */
void ThreadIndex::saveIndex(const Thread* thread)
{
    // If readNum == 0, delete idx file
    if (thread->readNum() == 0) {
        QFile::remove(m_indexPath);
    }
    KConfig config(m_indexPath);

    // save thread name
    setSubjectPrivate(config, thread->threadName());

    // save res number
    setResNumPrivate(config, thread->resNum());

    // save read number
    setReadNumPrivate(config, thread->readNum());

    // save view pos
    setViewPosPrivate(config, thread->viewPos());

    // save mark
    setMarkListPrivate(config, thread->markList());

    // save "cache"
    KUrl datUrl = getDatUrl(m_url);
    int num = (thread->viewPos() ? thread->viewPos() : thread->readNum());
    ThreadInfo::setReadNum(datUrl.prettyUrl(), num);
}


/*------------------------------------------------------------------*/

/* private */
QString ThreadIndex::getSubjectPrivate(const KConfig& config) const
{
    return config.group("").readEntry("Subject");
}

/* private */
void ThreadIndex::setSubjectPrivate(KConfig& config, const QString& str)
{
    config.group("").writeEntry("Subject", str);
}

/*-------*/

/* private */
int ThreadIndex::getResNumPrivate(KConfig& config) const
{
    int resNum = config.group("").readEntry("ResNum", 0);

    // use obsoleted "cache" file
    if (!resNum) {
        KUrl datUrl = getDatUrl(m_url);
        resNum = ThreadInfo::readNum(datUrl.prettyUrl());
        if (resNum)
            config.group("").writeEntry("ResNum", resNum);
    }

    return resNum;
}

/* private */
void ThreadIndex::setResNumPrivate(KConfig& config, int resNum)
{
    config.group("").writeEntry("ResNum", resNum);
}

/*-------*/

/* private */
int ThreadIndex::getReadNumPrivate(KConfig& config, bool checkCached) const
{
    // If cache does not exist, return 0
    if (checkCached) {
        Cache cache(m_url);
        QString path = cache.getPath();
        if (!QFile::exists(path)) {
            return 0;
        }
    }

    int readNum = config.group("").readEntry("ReadNum", 0);

    if (!readNum) {

        // use ViewPos instead of ReadNum.
        readNum = config.group("").readEntry("ViewPos", 0);

        // use obsoleted "cache" file
        if (!readNum) {
            KUrl datUrl = getDatUrl(m_url);
            readNum = ThreadInfo::readNum(datUrl.prettyUrl());
        }

        if (readNum)
            config.group("").writeEntry("ReadNum", readNum);
    }

    return readNum;
}

/* private */
void ThreadIndex::setReadNumPrivate(KConfig& config, int readNum)
{
    config.group("").writeEntry("ReadNum", readNum);
}

/*-------*/

/* private */
int ThreadIndex::getViewPosPrivate(const KConfig& config) const
{
    return config.group("").readEntry("ViewPos", 0);
}

/* private */
void ThreadIndex::setViewPosPrivate(KConfig& config, int viewPos)
{
    config.group("").writeEntry("ViewPos", viewPos);
}

/*-------*/

/* private */
QList<int> ThreadIndex::getMarkListPrivate(const KConfig& config) const
{
    QList<int> default_value;
    return config.group("").readEntry("Mark", default_value);
}

/* private */
void ThreadIndex::setMarkListPrivate(KConfig& config,
        const QList<int>& markList)
{
    config.group("").writeEntry("Mark", markList);
}
