/***************************************************************************
*   Copyright (C) 2004 by Kita Developers                                 *
*   ikemo@users.sourceforge.jp                                            *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
***************************************************************************/

#include "mainwindow.h"

#include <QtCore/QEvent>
#include <QtCore/QFile>
#include <QtGui/QBoxLayout>
#include <QtGui/QKeyEvent>
#include <QtGui/QSplitter>

#include <kaction.h>
#include <kactioncollection.h>
#include <kapplication.h>
#include <kconfig.h>
#include <kedittoolbar.h>
#include <kglobal.h>
#include <klineedit.h>
#include <klocale.h>
#include <kmenubar.h>
#include <kshortcutsdialog.h>
#include <kstandarddirs.h>
#include <kstatusbar.h>
#include <ktoggleaction.h>
#include <ktoolbar.h>
#include <kurl.h>
#include <kxmlguifactory.h>

#include "bbstabwidget.h"
#include "boardtabwidget.h"
#include "threadtabwidget.h"
#include "viewmediator.h"
#include "writetabwidget.h"
#include "libkita/abone.h"
#include "libkita/account.h"
#include "libkita/accountconfig.h"
#include "libkita/asciiart.h"
#include "libkita/colorconfig.h"
#include "libkita/datmanager.h"
#include "libkita/favoriteboards.h"
#include "libkita/favoritethreads.h"
#include "libkita/globalconfig.h"
#include "libkita/kita_misc.h"
#include "libkita/threadinfo.h"
#include "libkita/writeconfig.h"
#include "prefs/preferences.h"

using namespace Kita;

MainWindow::MainWindow() : KXmlGuiWindow(0), m_prefDialog(0)
{
    // accept dnd
    setAcceptDrops(true);

    // setup view, dock
    setupView();

    // then, setup our actions
    setupActions();

    // load cache
    loadCache();

    // load ascii art
    AsciiArtConfig::self()->readConfig();
    if (AsciiArtConfig::self()->asciiArtList().isEmpty()) {
        loadAsciiArt();
    }

    // load abone lists
    AboneConfig::self()->readConfig();
    if (AboneConfig::self()->aboneIDList().isEmpty()) {
        loadAboneIDList();
    }

    if (AboneConfig::self()->aboneNameList().isEmpty()) {
        loadAboneNameList();
    }

    if (AboneConfig::self()->aboneWordList().isEmpty()) {
        loadAboneWordList();
    }

    // and a status bar
    statusBar() ->show();

    // for compativility
    loadCompletion();

    AccountConfig::self()->readConfig();
    ColorConfig::self()->readConfig();
    GlobalConfig::self()->readConfig();
    WriteConfig::self()->readConfig();

    // apply the saved mainwindow settings, if any, and ask the mainwindow
    // to automatically save settings if changed: window size, toolbar
    // position, icon size, etc.
    setAutoSaveSettings();

    // set list font
    setFont(GlobalConfig::font());

    // allow the view to change the statusbar and caption
    connect(m_urlLine, SIGNAL(returnPressed()),
             SLOT(slotUrlLine()));

    QMenu* settingsPopup = static_cast<QMenu *>(factory() ->container("settings", this));
    connect(settingsPopup, SIGNAL(aboutToShow()),
             SLOT(settingsMenuAboutToShow()));

    // load favorite boards;
    loadFavoriteBoards();

    // load boad list
    {
        QString configPath = KStandardDirs::locateLocal("appdata", "board_list");
        if (QFile::exists(configPath)) {
            m_bbsTab->showBoardList();
            m_bbsTab->loadOpened();
        } else {
            m_bbsTab->updateBoardList();
        }
    }

    // load favorites
    loadFavorites();

    // update favorite list
    ViewMediator::getInstance()->updateFavoriteListView();

    if (AccountConfig::autoLogin()) {
        login();
    }
}

MainWindow::~MainWindow()
{
    AboneConfig::self()->writeConfig();

    AsciiArtConfig::self()->writeConfig();

    saveFavoriteBoards();

    saveFavorites();

    saveCache();

    saveMainWindowSettings(KGlobal::config().data()->group("MainWindow"));

    AccountConfig::self()->writeConfig();
    ColorConfig::self()->writeConfig();
    GlobalConfig::self()->writeConfig();
    WriteConfig::self()->writeConfig();

    DatManager::deleteAllDatInfo();
}

void MainWindow::load(const KUrl& url)
{
    setCaption(url.url());
}

void MainWindow::bookmark(const QString& datUrl, bool on)
{
    FavoriteThreads * favorite = FavoriteThreads::getInstance();

    on ? favorite->insert(datUrl) : favorite->remove(datUrl);
    saveFavorites();
    ViewMediator::getInstance()->updateFavoriteListView();
}

void MainWindow::login()
{
    if (Account::login(AccountConfig::userID(), AccountConfig::password())) {
        setMainStatus(i18nc("@info:status", "Login succeeded."));
    } else {
        setMainStatus(i18nc("@info:status", "Login failed."));
    }
}

void MainWindow::newToolbarConfig()
{
    // this slot is called when user clicks "Ok" or "Apply" in the toolbar editor.
    // recreate our GUI, and re-apply the settings (e.g. "text under icons", etc.)
    applyMainWindowSettings(KGlobal::config().data()->group("MainWindow"));
}

void MainWindow::optionsShowToolbar()
{
    // this is all very cut and paste code for showing/hiding the
    // toolbar
    if (m_toolbarAction->isChecked()) {
        toolBar() ->show();
    } else {
        toolBar() ->hide();
    }
}

void MainWindow::optionsShowStatusbar()
{
    // this is all very cut and paste code for showing/hiding the
    // statusbar
    if (m_statusbarAction->isChecked()) {
        statusBar() ->show();
    } else {
        statusBar() ->hide();
    }
}

void MainWindow::optionsConfigureKeys()
{
    KShortcutsDialog::configure(actionCollection());
}

void MainWindow::optionsConfigureToolbars()
{
    // use the standard toolbar editor
    KEditToolBar dlg(factory());
    connect(&dlg, SIGNAL(newToolbarConfig()),
             SLOT(newToolbarConfig()));
    dlg.exec();
}

void MainWindow::optionsPreferences()
{
    if (m_prefDialog == 0) {
        m_prefDialog = new Preferences(this);
        m_prefDialog->load();
        connect(m_prefDialog, SIGNAL(fontChanged(const QFont&)),
                 SLOT(setFont(const QFont&)));
        m_prefDialog->show();
    } else {
        m_prefDialog->show();
        m_prefDialog->raise();
    }
}

void MainWindow::settingsMenuAboutToShow()
{
    m_toolbarAction->setChecked(toolBar() ->isVisible());
    m_statusbarAction->setChecked(statusBar() ->isVisible());
}

void MainWindow::setFont(const QFont& font)
{
    m_boardTab->setFont(font);
    m_bbsTab->setFont(font);
}

void MainWindow::setUrl(const KUrl& url)
{
    m_urlLine->setText(url.url());
}

void MainWindow::slotEditCopy()
{
    QWidget * widget = KApplication::kApplication()->focusWidget();

    if (widget) {
        QKeyEvent e(QEvent::KeyPress, Qt::Key_C, Qt::ControlModifier);
        QApplication::sendEvent(widget, &e);
    }
}

void MainWindow::setMainStatus(const QString& statusStr)
{
    // display the text on the statusbar
    statusBar()->showMessage(statusStr);
}

void MainWindow::slotUrlLine()
{
    KUrl url = m_urlLine->text();
    KUrl datUrl = getDatUrl(url);
    m_threadTab->slotShowMainThread(datUrl);
}

//
// private method
//

void MainWindow::setupActions()
{
    KStandardAction::quit(this, SLOT(close()), actionCollection());
    KStandardAction::copy(this, SLOT(slotEditCopy()), actionCollection());

    setStandardToolBarMenuEnabled(true);

    m_toolbarAction
        = new KToggleAction(i18nc("@action:inmenu", "Show Toolbar"), this);
    actionCollection()->addAction("toolBar", m_toolbarAction);
    connect(m_toolbarAction, SIGNAL(toggled(bool)),
            SLOT(optionsShowToolbar()));

    m_statusbarAction = KStandardAction::showStatusbar(this,
                        SLOT(optionsShowStatusbar()),
                        actionCollection());

    m_urlLine = new KLineEdit;
    KAction* urlLineAction = new KAction(this);
    urlLineAction->setDefaultWidget(m_urlLine);
    urlLineAction->setText(i18nc("@action:intoolbar", "Location Bar"));
    actionCollection()->addAction("url_line_action", urlLineAction);
    connect(urlLineAction, SIGNAL(triggered()), this, SLOT(slotUrlLine()));

    KStandardAction::keyBindings(this, SLOT(optionsConfigureKeys()),
            actionCollection());
    KStandardAction::configureToolbars(this, SLOT(optionsConfigureToolbars()),
            actionCollection());
    KStandardAction::preferences(this, SLOT(optionsPreferences()),
            actionCollection());

    KAction* loadBoardAction
        = actionCollection()->addAction("load_board_list");
    loadBoardAction->setText(i18nc("@action:inmenu", "Load Board List"));
    connect(loadBoardAction, SIGNAL(triggered()),
            m_bbsTab, SLOT(updateBoardList()));

    KAction* loginAction = actionCollection()->addAction("login");
    loginAction->setText(i18nc("@action:inmenu Login menu", "Login"));
    connect(loginAction, SIGNAL(triggered()), m_bbsTab, SLOT(login()));

    setXMLFile("kitaui.rc");
    KXmlGuiWindow::createGUI();
    factory()->addClient(m_bbsTab);
    factory()->addClient(m_boardTab);
    factory()->addClient(m_threadTab);
    factory()->addClient(m_writeTab);
}

void MainWindow::setupView()
{
    ViewMediator::getInstance()->setMainWindow(this);

    QWidget* mainWidget = new QWidget(this);

    QBoxLayout* mainLayout = new QVBoxLayout(mainWidget);
    QSplitter* hsplit = new QSplitter(mainWidget);
    mainLayout->addWidget(hsplit);

    m_bbsTab = new BBSTabWidget(hsplit);

    QSplitter* vsplit = new QSplitter(Qt::Vertical, hsplit);

    m_boardTab = new BoardTabWidget(vsplit);
    ViewMediator::getInstance()->setBoardTabWidget(m_boardTab);

    m_threadTab = new ThreadTabWidget(vsplit);
    ViewMediator::getInstance()->setThreadTabWidget(m_threadTab);

    hsplit->setSizes(QList<int>() << 100 << 500);
    vsplit->setSizes(QList<int>() << 200 << 300);

    setCentralWidget(mainWidget);

    // write dock
    m_writeTab = new WriteTabWidget(0);
    ViewMediator::getInstance()->setWriteTabWidget(m_writeTab);
}

void MainWindow::loadCache()
{
    ThreadInfo * cache = ThreadInfo::getInstance();
    QString cacheConfigPath = KStandardDirs::locateLocal("appdata", "cache");
    QFile file(cacheConfigPath);
    if (file.open(QIODevice::ReadOnly)) {
        QDataStream stream(&file);
        stream >> *cache;
    }
}

void MainWindow::saveCache()
{
    ThreadInfo * cache = ThreadInfo::getInstance();
    QString cacheConfigPath = KStandardDirs::locateLocal("appdata", "cache");
    QFile file(cacheConfigPath);
    if (file.open(QIODevice::WriteOnly)) {
        QDataStream stream(&file);
        stream << *cache;
    }
}

void MainWindow::loadFavorites()
{
    QString favoritesConfigPath = KStandardDirs::locateLocal("appdata", "favorites.xml");
    QFile file(favoritesConfigPath);
    if (file.open(QIODevice::ReadOnly)) {
        QTextStream stream(&file);
        stream.setCodec("UTF-8");

        QString xml = stream.readAll();
        FavoriteThreads::readFromXML(xml);
    }
}

void MainWindow::saveFavorites()
{
    QString favoritesConfigPath
        = KStandardDirs::locateLocal("appdata", "favorites.xml");
    QFile file(favoritesConfigPath);
    if (file.open(QIODevice::WriteOnly)) {
        QTextStream stream(&file);
        stream.setCodec("UTF-8");
        stream << FavoriteThreads::getInstance()->toXML();
    }
}

void MainWindow::loadFavoriteBoards()
{
    QString configPath = KStandardDirs::locateLocal("appdata", "favorite_boards.xml");
    QFile file(configPath);
    if (file.open(QIODevice::ReadOnly)) {
        QTextStream stream(&file);
        stream.setCodec("UTF-8");

        QString xml = stream.readAll();
        FavoriteBoards::readFromXML(xml);
    }
}

void MainWindow::saveFavoriteBoards()
{
    QString configPath = KStandardDirs::locateLocal("appdata", "favorite_boards.xml");
    QFile file(configPath);
    if (file.open(QIODevice::WriteOnly)) {
        QTextStream stream(&file);
        stream.setCodec("UTF-8");
        stream << FavoriteBoards::toXML();
    }
}

void MainWindow::loadCompletion()
{
    QString configPath = KStandardDirs::locateLocal("appdata", "completion");
    KConfig config(configPath);

    GlobalConfig::setNameCompletionList(
            config.group("").readEntry("name", QStringList()));
}

void MainWindow::loadAsciiArt()
{
    QString configPath = KStandardDirs::locateLocal("appdata", "asciiart");
    QFile file(configPath);
    if (file.open(QIODevice::ReadOnly)) {
        QTextStream stream(&file);
        stream.setCodec("UTF-8");

        QStringList list;
        QString str;

        while (!(str = stream.readLine()).isEmpty()) {
            if (! str.isEmpty()) {
                list << str;
            }
        }
        AsciiArtConfig::setAsciiArtList(list);
    }
}

void MainWindow::loadAboneIDList()
{
    QString configPath = KStandardDirs::locateLocal("appdata", "abone_id");
    QFile file(configPath);
    if (file.open(QIODevice::ReadOnly)) {
        QTextStream stream(&file);
        stream.setCodec("UTF-8");

        QStringList list;
        QString str;

        while (!(str = stream.readLine()).isEmpty()) {
            if (! str.isEmpty()) {
                list << str;
            }
        }
        AboneConfig::setAboneIDList(list);
    }
}

void MainWindow::loadAboneNameList()
{
    QString configPath = KStandardDirs::locateLocal("appdata", "abone_name");
    QFile file(configPath);
    if (file.open(QIODevice::ReadOnly)) {
        QTextStream stream(&file);
        stream.setCodec("UTF-8");

        QStringList list;
        QString str;

        while (!(str = stream.readLine()).isEmpty()) {
            if (! str.isEmpty()) {
                list << str;
            }
        }
        AboneConfig::setAboneNameList(list);
    }
}

void MainWindow::loadAboneWordList()
{
    QString configPath = KStandardDirs::locateLocal("appdata", "abone_word");
    QFile file(configPath);
    if (file.open(QIODevice::ReadOnly)) {
        QTextStream stream(&file);
        stream.setCodec("UTF-8");

        QStringList list;
        QString str;

        while (!(str = stream.readLine()).isEmpty()) {
            if (! str.isEmpty()) {
                list << str;
            }
        }
        AboneConfig::setAboneWordList(list);
    }
}
