/***************************************************************************
*   Copyright (C) 2003 by Hideki Ikemoto                                  *
*   ikemo@wakaba.jp                                                       *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
***************************************************************************/

#include "threadlistview.h"

#include <QtGui/QClipboard>
#include <QtGui/QContextMenuEvent>
#include <QtGui/QHeaderView>

#include <kaction.h>
#include <kmenu.h>
#include <krun.h>
#include <kurl.h>

#include "threadlistheaderview.h"
#include "threadlistviewitem.h"
#include "ui_threadproperty.h"
#include "viewmediator.h"
#include "libkita/cache.h"
#include "libkita/datmanager.h"
#include "libkita/favoritethreads.h"
#include "libkita/kita_misc.h"
#include "libkita/threadindex.h"
#include "libkita/threadinfo.h"

using namespace Kita;

ThreadListView::ThreadListView(QWidget* parent)
        : QWidget(parent)
{
    setupUi(this);
    hideButton->setIcon(SmallIcon("view-filter"));
    reloadButton->setIcon(SmallIcon("view-refresh"));
    closeButton->setIcon(SmallIcon("tab-close"));

    ThreadListHeaderView* header = new ThreadListHeaderView;

    subjectList->setAlternatingRowColors(true);
    subjectList->setColumnCount(ColumnEnd - ColumnBegin + 1);
    subjectList->setHorizontalHeader(header);
    subjectList->setShowGrid(false);
    subjectList->setSortingEnabled(true);
    subjectList->verticalHeader()->setVisible(false);
    QStringList labels;
    for (int i = ColumnBegin; i <= ColumnEnd; i++) {
        labels << i18nc("@title:column",
                ThreadListHeaderView::attribute[i].labelName);
        if (!ThreadListHeaderView::attribute[i].showDefault) {
            hideColumn(i);
        }
    }
    subjectList->setHorizontalHeaderLabels(labels);
    subjectList->setSelectionBehavior(QAbstractItemView::SelectRows);
    subjectList->setSelectionMode(QAbstractItemView::SingleSelection);

    connect(searchCombo, SIGNAL(activated(int)),
             SLOT(slotSearchButton()));
    connect(searchCombo, SIGNAL(textChanged(const QString&)),
             SLOT(slotSearchButton()));
    connect(hideButton, SIGNAL(toggled(bool)),
             SLOT(slotHideButton(bool)));
    connect(subjectList, SIGNAL(itemClicked(QTableWidgetItem*)),
             SLOT(slotItemClicked(QTableWidgetItem*)));

    m_openWithBrowserAct
        = new KAction(i18nc("@action:inmenu", "Open with Web Browser") , this);
    m_copyUrlAct = new KAction(i18nc("@action:inmenu", "Copy URL"), this);
    m_copyTitleAndUrlAct
        = new KAction(i18nc("@action:inmenu", "Copy title and URL"), this);
    m_favoritesAct = new KAction(this);
    m_deleteLogAct = new KAction(i18nc("@action:inmenu", "Delete Log"), this);
    m_deleteLogAct->setVisible(false);
    m_propertyAct = new KAction(i18nc("@action:inmenu", "Property"), this);
}

ThreadListView::~ThreadListView() {}

void ThreadListView::slotSearchButton()
{
    insertSearchCombo();
    QStringList list = parseSearchQuery(searchCombo->currentText());

    if (list.isEmpty()) {
        clearSearch();
    } else if (list != m_prevquery) {
        searchNew(list);
        hideButton->setDown(true);
        slotHideButton(true);
    } else {
        searchNext(list);
    }
}

void ThreadListView::insertSearchCombo()
{
    for (int count = 0; count < searchCombo->count(); ++count) {
        if (searchCombo->itemText(count) == searchCombo->currentText()) {
            return;
        }
    }
    searchCombo->addItem(searchCombo->currentText());
}

void ThreadListView::searchNext(const QStringList &query)
{
    Q_ASSERT(query == m_prevquery);
    Q_ASSERT(m_nextHitIndex < static_cast<unsigned>(m_hitList.size()));

    QTableWidgetItem* item = m_hitList[m_nextHitIndex];
    subjectList->scrollToItem(item);

    m_nextHitIndex++;
    if (m_nextHitIndex >= static_cast<unsigned>(m_hitList.size()))
        m_nextHitIndex = 0;
}

void ThreadListView::searchNew(const QStringList &query)
{
    m_hitList.clear();
    m_nextHitIndex = 0;
    m_prevquery = query;
    for (int i = 0, j = subjectList->rowCount(); i < j; i++) {
        subjectList->item(i, ColumnIcon)->setIcon(QIcon());
        for (int k = 0, l = query.count(); k < l; k++) {
            if (subjectList->item(i, ColumnSubject)->text()
                    .contains(query[k], Qt::CaseInsensitive)) {
                subjectList->item(i, ColumnIcon)
                    ->setIcon(SmallIcon("edit-find"));
                m_hitList.append(subjectList->item(i, 0));
                break;
            }
        }
    }
}

void ThreadListView::clearSearch()
{
    for (int i = 0, j = subjectList->rowCount(); i < j; i++) {
        subjectList->item(i, ColumnIcon)->setIcon(QIcon());
        subjectList->showRow(i);
    }
}

void ThreadListView::slotHideButton(bool on)
{
    for (int i = 0, j = subjectList->rowCount(); i < j; i++) {
        if (on && subjectList->item(i, ColumnIcon)->icon().isNull()) {
            subjectList->hideRow(i);
        } else {
            subjectList->showRow(i);
        }
    }
}

void ThreadListView::slotItemClicked(QTableWidgetItem* item)
{
    if (!item) return;
    KUrl datUrl = subjectList->item(item->row(), ColumnDatUrl)->text();
    ViewMediator::getInstance()->openThread(datUrl);
}

void ThreadListView::hideColumn(int col)
{
    subjectList->hideColumn(col);
}

void ThreadListView::showColumn(int col)
{
    subjectList->showColumn(col);
}

void ThreadListView::contextMenuEvent(QContextMenuEvent *event)
{
    QTableWidgetItem* item = subjectList->itemAt(
            subjectList->viewport()->mapFromGlobal(event->globalPos()));
    if (item == 0) {
        return;
    }

    QString datUrl = subjectList->item(item->row(), ColumnDatUrl)->text();
    DatManager datManager(datUrl);
    QString threadUrl = datManager.threadUrl();

    // create popup menu.
    KMenu popup(0);

    popup.addAction(m_openWithBrowserAct);
    popup.addAction(m_copyUrlAct);
    popup.addAction(m_copyTitleAndUrlAct);

    bool isFavorites = FavoriteThreads::getInstance()->contains(datUrl);
    m_favoritesAct->setText(isFavorites
        ? i18nc("@action:inmenu", "Remove from Favorites")
        : i18nc("@action:inmenu", "Add to Favorites"));
    popup.addAction(m_favoritesAct);

    if (datManager.getReadNum()) {
        popup.addSeparator();
        popup.addAction(m_deleteLogAct);
    }
    popup.addSeparator();
    popup.addAction(m_propertyAct);

    // exec popup menu.
    QClipboard* clipboard = QApplication::clipboard();

    QAction* action = popup.exec(QCursor::pos());
    if (!action) {
        return;
    }
    if (action == m_openWithBrowserAct) {
        KRun::runUrl(threadUrl, "text/html", this);
    } else if (action == m_copyUrlAct) {
        clipboard->setText(threadUrl, QClipboard::Clipboard);
        clipboard->setText(threadUrl, QClipboard::Selection);
    } else if (action == m_copyTitleAndUrlAct) {
        QString cliptxt = datManager.threadName() + '\n' + threadUrl;
        clipboard->setText(cliptxt , QClipboard::Clipboard);
        clipboard->setText(cliptxt , QClipboard::Selection);
    } else if (action == m_favoritesAct) {
        ViewMediator::getInstance()->bookmark(datUrl, !isFavorites);
    } else if (action == m_deleteLogAct) {
        deleteLog(threadUrl);
    } else if (action == m_propertyAct) {
            QWidget* widget = new QWidget;
            Ui::ThreadProperty* propertyWidget
                = new Ui::ThreadProperty;
            propertyWidget->setupUi(widget);
            propertyWidget->threadUrlLabel->setText(threadUrl);
            propertyWidget->datUrlLabel->setText(datUrl);

            propertyWidget->threadNameLabel
                ->setText(datManager.threadName());
            propertyWidget->cachePathLabel
                ->setText(Cache(datUrl).getPath());
            propertyWidget->indexPathLabel
                ->setText(Cache(datUrl).getIndexPath());
            propertyWidget->resNumLabel
                ->setText(QString::number(datManager.getResNum()));
            propertyWidget->readNumLabel
                ->setText(QString::number(datManager.getReadNum()));
            propertyWidget->viewPosLabel
                ->setText(QString::number(datManager.getViewPos()));

            ThreadIndex threadIndex(datUrl);
            propertyWidget->idx_threadNameWithIndexLabel
                ->setText(threadIndex.getSubject());
            propertyWidget->idx_resNumLabel
                ->setText(QString::number(threadIndex.getResNum()));
            propertyWidget->idx_readNumLabel
                ->setText(QString::number(threadIndex.getReadNum()));
            propertyWidget->idx_viewPosLabel
                ->setText(QString::number(threadIndex.getViewPos()));

            propertyWidget->cache_readNumLabel
                ->setText(QString::number(ThreadInfo::readNum(datUrl)));
            widget->show();
    }
}

// dummy function. reimplemented in BoardView
void ThreadListView::deleteLog(const KUrl&) {}
