/***************************************************************************
*   Copyright (C) 2004 by Kita Developers                                 *
*   ikemo@users.sourceforge.jp                                            *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
***************************************************************************/

#include "threadtabwidget.h"

#include <QtGui/QApplication>
#include <QtGui/QClipboard>
#include <QtGui/QContextMenuEvent>

#include <kaction.h>
#include <kactioncollection.h>
#include <klocale.h>
#include <kmenu.h>
#include <kstandardaction.h>
#include <kstandardshortcut.h>

#include "threadview.h"
#include "viewmediator.h"
#include "libkita/datmanager.h"
#include "libkita/globalconfig.h"
#include "libkita/kita_misc.h"

using namespace Kita;

ThreadTabWidget::ThreadTabWidget(QWidget* parent) : TabWidgetBase(parent)
{
    setXMLFile("threadtabwidgetui.rc");

    m_actionList = actionCollection()->actions();

    setupActions();

    // default view
    createView(i18nc("@title:tab", "Thread"));
}


ThreadTabWidget::~ThreadTabWidget() {}

/* show "Main thread" view */  /* public slots */
void ThreadTabWidget::slotShowMainThread(const KUrl& url)
{
    QString refstr;
    KUrl datUrl = getDatUrl(url, refstr);
    QString threadName = DatManager(datUrl).threadName();
    int jumpNum = 0;

    int viewMode = VIEWMODE_MAINVIEW;
    ThreadView* currentView = isThreadView(currentWidget());
    if (currentView) viewMode = currentView->getViewMode();

    if (!refstr.isEmpty()) {
        int i = refstr.indexOf("-");
        if (i != -1) jumpNum = refstr.left(i).toInt();
        else jumpNum = refstr.toInt();
    }

    ThreadView* view = findMainView(datUrl);

    if (view) {

        setCurrentWidget(view);

        if (view->threadUrl().isEmpty()) {
            // Show on the default view
            view->showThread(datUrl, jumpNum);
        } else {
            // The view is already shown
            // TODO: jump to jumpNum after reloading

            view->slotReloadButton(jumpNum);
        }

    } else {
        ThreadView * newView = createView(threadName);

        if (newView) {
            newView->showThread(datUrl, jumpNum);
            setCurrentWidget(newView);
        }
    }

    slotUpdateThreadTab(datUrl);
}

/* close "all" views which URL is url. */ /* public slot */
void ThreadTabWidget::slotCloseThreadTab(const KUrl& url)
{
    int max = count();
    if (max == 0) return ;
    KUrl datUrl = getDatUrl(url);

    int i, i2;
    i = i2 = 0;
    while (i < max) {
        ThreadView * view = isThreadView(widget(i));
        if (view) {
            if (view->datUrl() == datUrl) {
                slotCloseTab(i2);
                i2--;
            }
        }
        i++; i2++;
    }
}



/* create ThreadView */   /* private */
ThreadView* ThreadTabWidget::createView(const QString& label)
{
    ThreadView * view = new ThreadView(this);
    if (view) {
        addTab(view, label);
    }

    return view;
}

/* private */
ThreadView* ThreadTabWidget::findMainView(const KUrl& url)
{
    KUrl datUrl = getDatUrl(url);

    int max = count();
    if (max == 0) return 0;
    int i = 0;

    while (i < max) {
        ThreadView * view = isThreadView(widget(i));
        if (view) {
            if (view->getViewMode() == VIEWMODE_MAINVIEW) {

                if (view->datUrl() == datUrl
                        || view->datUrl().isEmpty()  // default view
                  ) return view;
            }
        }
        i++;
    }

    return 0;
}


/* private */
ThreadView* ThreadTabWidget::isThreadView(QWidget* w)
{
    return qobject_cast<ThreadView*>(w);
}


/* private slots */
void ThreadTabWidget::slotUpdateThreadTab(const KUrl& url)
{
    KUrl datUrl = getDatUrl(url);

    ThreadView * view = findMainView(datUrl);
    if (view) {
        QString threadName = DatManager(datUrl).threadName();

        setTabText(indexOf(view), threadName);
        setTabToolTip(indexOf(view), threadName);
    }
}

void ThreadTabWidget::slotFontChanged()
{
    QFont font = GlobalConfig::threadFont();
    setFont(font);
}

/* protected */ /* virtual */
void ThreadTabWidget::deleteWidget(QWidget* w)
{
    TabWidgetBase::deleteWidget(w);

    if (count() == 0) {
        ViewMediator::getInstance()->setMainCaption(QString());
        ViewMediator::getInstance()->setMainStatus(QString());
        ViewMediator::getInstance()->setMainUrlLine(KUrl());

        // default view
        ThreadView * threadView = createView(i18nc("@title:tab", "Thread"));

        if (threadView) {
            setCurrentWidget(threadView);
        }
    }
}



/*--------------------------------*/
/* ThreadView actions */


/* private */
void ThreadTabWidget::setupActions()
{
    KStandardAction::copy(this, SLOT(slotCopyText()),
            actionCollection());
    KStandardAction::find(this, SLOT(slotFocusSearchCombo()),
            actionCollection());
    KStandardAction::findNext(this, SLOT(slotSearchNext()),
            actionCollection());
    KStandardAction::findPrev(this, SLOT(slotSearchPrev()),
            actionCollection());

    KAction* gobackAction = actionCollection()->addAction("threadview_goback");
    gobackAction->setText(i18nc("@action:inmenu", "Go Back Anchor"));
    gobackAction->setShortcut(Qt::Key_Backspace);
    gobackAction->setIcon(SmallIcon("arrow-left-double"));
    connect(gobackAction, SIGNAL(triggered()), this, SLOT(slotGobackAnchor()));

    KAction* gotoheaderAction
        = actionCollection()->addAction("threadview_gotoheader");
    gotoheaderAction->setText(i18nc("@action:inmenu", "Start"));
    gotoheaderAction->setShortcut(Qt::Key_Home);
    gotoheaderAction->setIcon(SmallIcon("go-up"));
    connect(gotoheaderAction, SIGNAL(triggered()),
            this, SLOT(slotGotoHeader()));

    KAction* gotofooterAction
        = actionCollection()->addAction("threadview_gotofooter");
    gotofooterAction->setText(i18nc("@action:inmenu", "End"));
    gotofooterAction->setShortcut(Qt::Key_End);
    gotofooterAction->setIcon(SmallIcon("go-down"));
    connect(gotofooterAction, SIGNAL(triggered()), this, SLOT(slotGotoFooter()));

    KAction* reload_action
        = actionCollection()->addAction("threadview_reload");
    reload_action->setText(i18nc("@action:button", "Reload"));
    reload_action->setShortcut(KStandardShortcut::reload());
    reload_action->setIcon(SmallIcon("view-refresh"));
    connect(reload_action, SIGNAL(triggered()), this, SLOT(slotReloadButton()));

    KAction* stopAction = actionCollection()->addAction("threadview_stop");
    stopAction->setText(i18nc("@action:inmenu", "Stop"));
    stopAction->setShortcut(Qt::Key_Escape);
    stopAction->setIcon(SmallIcon("process-stop"));
    connect(stopAction, SIGNAL(triggered()), this, SLOT(slotStopLoading()));

    KAction* deleteAction = actionCollection()->addAction("threadview_delete");
    deleteAction->setText(i18nc("@action:button", "Delete"));
    deleteAction->setShortcut(Qt::Key_Delete);
    deleteAction->setIcon(SmallIcon("trash-empty"));
    connect(deleteAction, SIGNAL(triggered()), this, SLOT(slotDeleteButtonClicked()));
}



/* ThreadView actions */

/* copy selected text (Ctrl+C) */ /* public slot */
void ThreadTabWidget::slotCopyText()
{
    QWidget * w = currentWidget();
    ThreadView * view = isThreadView(w);
    if (view) {
        QClipboard * clipboard = QApplication::clipboard();
        QString text = view->selectedText();
        clipboard->setText(text, QClipboard::Clipboard);
    }
}

/* public slot  */
void ThreadTabWidget::slotFocusSearchCombo()
{
    ThreadView * view = isThreadView(currentWidget());
    if (view) view->focusSearchCombo();
}


/* public slot  */
void ThreadTabWidget::slotSearchNext()
{
    ThreadView * view = isThreadView(currentWidget());
    if (view) view->slotSearchNext();
}


/* public slot  */
void ThreadTabWidget::slotSearchPrev()
{
    ThreadView * view = isThreadView(currentWidget());
    if (view) view->slotSearchPrev();
}


/* public slot  */
void ThreadTabWidget::slotGobackAnchor()
{
    ThreadView * view = isThreadView(currentWidget());
    if (view) view->slotGobackAnchor();
}


/* public slot  */
void ThreadTabWidget::slotGotoHeader()
{
    ThreadView * view = isThreadView(currentWidget());
    if (view) view->slotGotoHeader();
}


/* public slot  */
void ThreadTabWidget::slotGotoFooter()
{
    ThreadView * view = isThreadView(currentWidget());
    if (view) view->slotGotoFooter();
}


/* public slot  */
void ThreadTabWidget::slotReloadButton()
{
    ThreadView * view = isThreadView(currentWidget());
    if (view) view->slotReloadButton();
}


/* public slot  */
void ThreadTabWidget::slotStopLoading()
{
    ThreadView * view = isThreadView(currentWidget());
    if (view) view->slotStopLoading();
}


/* public slot  */
void ThreadTabWidget::slotDeleteButtonClicked()
{
    ThreadView * view = isThreadView(currentWidget());
    if (view) view->slotDeleteButtonClicked();
}

void ThreadTabWidget::contextMenuEvent(QContextMenuEvent* e)
{
    KMenu popup;
    QAction* act;
    foreach (act, m_actionList) {
         popup.addAction(act);
    }
    popup.exec(e->globalPos());
//    KActionCollection * collection = client->actionCollection();
//    KAction* action;
//    action = new KAction(i18n("goback anchor"), SmallIcon("idea"), KShortcut(), m_threadPart, SLOT(gobackAnchor()), collection, "goback_anchor");
//    emit popupMenu(client, global, url, mimeType, mode); // TODO
}
