/***************************************************************************
 *   Copyright (C) 2004 by Kita Developers                                 *
 *   ikemo@users.sourceforge.jp                                            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include "writeview.h"

#include <QtCore/QFile>
#include <QtCore/QTextCodec>

#include <kmessagebox.h>
#include <kstandarddirs.h>
#include <kio/job.h>

#include "viewmediator.h"
#include "writetabwidget.h"
#include "libkita/account.h"
#include "libkita/accountconfig.h"
#include "libkita/asciiart.h"
#include "libkita/boarddatabase.h"
#include "libkita/datmanager.h"
#include "libkita/flashcgi.h"
#include "libkita/globalconfig.h"
#include "libkita/jbbs.h"
#include "libkita/kita_misc.h"
#include "libkita/kita-utf8.h"
#include "libkita/k2ch.h"
#include "libkita/machibbs.h"
#include "libkita/postdata.h"
#include "libkita/writeconfig.h"

using namespace Kita;

/*

Call setMessage() to set message later.
See also WriteTabWidget::slotShowWriteView().

Call slotPostMessage() to post the message.

*/
WriteView::WriteView(WriteTabWidget* parent, const KUrl& url)
        : QWidget(parent)
{
    setupUi(this);
    m_datUrl = getDatUrl(url);
    BoardDatabase db(m_datUrl);
    m_bbstype = db.type();
    m_bbscgi = getWriteUrl(m_datUrl);
    m_parent = parent;

    initUI();
}

void WriteView::initUI()
{
    // connect signals
    connect(buttonOk, SIGNAL(clicked()),
             SLOT(slotPostMessage()));

    connect(buttonCancel, SIGNAL(clicked()),
             SLOT(slotCancel()));

    connect(faceCombo, SIGNAL(activated(const QString&)),
             bodyText, SLOT(insertPlainText(const QString&)));

    connect(bodyText, SIGNAL(textChanged()),
             SLOT(slotBodyTextChanged()));

    // setup preview view
    qtw->setCurrentIndex(0);

    // setup labels and edit lines
    QFont font = GlobalConfig::threadFont();
    bodyText->setFont(font);
    bodyText->setTabChangesFocus(true);

    BoardDatabase db(m_datUrl);
    boardNameLabel->setText(db.boardName());

    // setup name field.
    nameLine->setText(WriteConfig::defaultName());
    QStringList compList = GlobalConfig::self()->nameCompletionList();
    nameLine->completionObject()->setItems(compList);

    // setup mail field & 'sage' checkbox
    if (WriteConfig::defaultSage()) {
        mailLine->setText("sage");
        sageBox->setChecked(true);
    } else {
        mailLine->setText(WriteConfig::defaultMail());
    }
    m_mailswap = "";

    // setup 'be' checkbox
    QRegExp host_2ch(".+\\.2ch\\.net");
    if (host_2ch.indexIn(m_bbscgi.host()) != -1
        && AccountConfig::beMailAddress().length() > 0
        && AccountConfig::beAuthCode().length() > 0) {
        beBox->setChecked(true);
    }

    // setup AA
    faceCombo->clear();
    faceCombo->setFont(GlobalConfig::threadFont());
    faceCombo->addItem("");
    QStringList list = AsciiArtConfig::asciiArtList();
    QStringList::iterator it;
    for (it = list.begin(); it != list.end(); ++it) {
        faceCombo->addItem(*it);
    }

    // init thread name field.
    threadNameLine->setText(DatManager(m_datUrl).threadName());
    threadNameLine->setReadOnly(true);
    threadNameLine->setFrame(false);
    threadNameLine->setFocusPolicy(Qt::NoFocus);
}

WriteView::~WriteView()
{
}

/* public */
void WriteView::setMessage(const QString& bodyStr)
{
    bodyText->clear();
    bodyText->insertPlainText(bodyStr);
    bodyText->setFocus();
}


/* public */
void WriteView::insertMessage(const QString& str)
{
    bodyText->insertPlainText(str);
    bodyText->setFocus();
}


/* public information */
const KUrl WriteView::datUrl() const
{
    return m_datUrl;
}

const QString WriteView::threadName() const
{
    return threadNameLine->text();
}

const QString WriteView::boardId() const
{
    BoardDatabase db(m_datUrl);
    return db.boardId();
}

const QString WriteView::boardName() const
{
    BoardDatabase db(m_datUrl);
    return db.boardName();
}
/* public slot */ /* virtual */
void WriteView::setFocus()
{
    bodyText->setFocus();
}

bool WriteView::checkFields()
{
    if (body().length() == 0) {
        return false;
    } else {
        return true;
    }
}

QString WriteView::buildPostMessage()
{
    PostData postData(m_datUrl, name(), mail(), body(), boardId());
    QString postStr;
    switch (m_bbstype) {
    case Board_JBBS:
        postStr = JBBS().buildPostStr(postData);
        break;

    case Board_FlashCGI:
        postStr = FlashCGI().buildPostStr(postData);
        break;

    case Board_MachiBBS:
        postStr = MachiBBS().buildPostStr(postData);
        break;

    default:
        postStr = K2ch().buildPostStr(postData);
        break;
    }

    return postStr;
}

/* call this slot to post the message. */ /* public slot */
void WriteView::slotPostMessage()
{
    if (!checkFields()) return;

    QString name = nameLine->text();
    QStringList list = GlobalConfig::nameCompletionList();
    list.append(name);
    GlobalConfig::setNameCompletionList(list);

    // build post message
    QString postStr = buildPostMessage();

    // referrer
    BoardDatabase db(m_datUrl);
    QString refStr = db.boardUrl();

    m_array.clear();

    KIO::TransferJob* job = KIO::http_post(m_bbscgi, postStr.toUtf8());
    job->addMetaData("content-type", "Content-type: application/x-www-form-urlencoded");
    job->addMetaData("referrer", refStr);

    // 2ch.net cookie modify
    if (m_bbstype == Board_2ch && beBox->isChecked()) {
        QString cookie = "Cookie: ";
        QString BeMailAddress = AccountConfig::beMailAddress();
        QString BeAuthCode = AccountConfig::beAuthCode();
        if (BeMailAddress.length() > 0 && BeAuthCode.length() > 0) {
            cookie += "DMDM=" + BeMailAddress + "; ";
            cookie += "MDMD=" + BeAuthCode + "; ";
        }
        job->addMetaData("customHTTPHeader", cookie);
    }

    connect(job, SIGNAL(data(KIO::Job*, const QByteArray&)),
             this, SLOT(slotRecieveData(KIO::Job*, const QByteArray&)));

    // slotPostFinished() is called when done.
    connect(job, SIGNAL(result(KJob*)),
             this, SLOT(slotPostFinished(KJob*)));

}


/* public slot */
void WriteView::slotCancel()
{
    if (body().length() == 0) {
        m_parent->slotCloseCurrentTab();
        return ;
    }

    if (KMessageBox::warningYesNo(this,
                i18n("If you close this dialog, you will lose text.\n"
                "Do you want to close?"), "Kita") == KMessageBox::No) return;
    setMessage(QString());
    m_parent->slotCloseCurrentTab();
}


/* public slot */
void WriteView::slotEnableWriting(bool enable)
{
    buttonOk->setEnabled(enable);
}


/* see also slotPostMessage() */ /* private slot */
void WriteView::slotRecieveData(KIO::Job*, const QByteArray& data)
{
    m_array.append(data.data());
}

void WriteView::processPostFinished()
{
    QString response;
    QString ckstr = QString::fromUtf8(KITAUTF8_WRITECOOKIE);

    // x-euc-jp & euc-jp
    if (m_array.contains("euc-jp")) {
        response = QTextCodec::codecForName("eucJP")->toUnicode(m_array);
    } else {
        response = QTextCodec::codecForName("sjis")->toUnicode(m_array);
    }

    int retcode = resultCode(response);
    switch (retcode) {

    case K2ch_True:

        // save log
        logPostMessage();

        // clear message
        setMessage(QString());

        // reload thread
        ViewMediator::getInstance()->openThread(m_datUrl);

        m_parent->slotCloseCurrentTab();

        break;

    case K2ch_Unknown:
    case K2ch_False:
    case K2ch_Check:
    case K2ch_Error:

        KMessageBox::error(0, resultMessage(response), resultTitle(response));

        break;

        // eat cookie, then re-post message
    case K2ch_Cookie:

        if (KMessageBox::questionYesNo(0,
            QString::fromUtf8(KITAUTF8_WRITECOOKIEMSG), resultTitle(response))
                == KMessageBox::Yes) {
            slotPostMessage();
        }

        break;

    default:
        break;
    }
}

/* This slot is called when posting is done. */
/* see also slotPostMessage()                */ /* private slot */
void WriteView::slotPostFinished(KJob*)
{
    processPostFinished();
}

int WriteView::getWriteResNum()
{
    return DatManager(m_datUrl).getReadNum() + 1;
}

/* private slot */
bool WriteView::slotBodyTextChanged()
{
    QString text = bodyText->toPlainText();
    int lines = text.count('\n') + 1;
    int length
        = QTextCodec::codecForName("Shift_JIS")->fromUnicode(body()).length();

    // replace '\n' -> " <br> ", '>' -> "&lt;", and etc.
    length += ((body().count('\n')) * 5
                + (body().count('>')) * 3
                + (body().count('<')) * 3
             );

    QString str;
    str = QString::number(lines);
    str += " | " + QString::number(length);
    lengthLabel->setText(str);

    return true;
}

/* save post log */ /* private */
void WriteView::logPostMessage()
{
    QString threadUrl = DatManager(m_datUrl).threadUrl();

    QDateTime now = QDateTime::currentDateTime();
    QString logPath = KStandardDirs::locateLocal("appdata", "log.txt");

    QFile file(logPath);
    if (file.open(QIODevice::WriteOnly | QIODevice::Append)) {
        QTextStream stream(&file);
        stream.setCodec("UTF-8");

        stream << "Date   : " << now.toString("yyyy/MM/dd hh:mm:ss") << endl; // current time
        stream << "Board  : " << boardName() << endl;
        stream << "Thread : " << threadName() << endl;
        stream << "URL    : " << threadUrl << endl;
        stream << "Name   : " << name() << endl;
        stream << "Mail   : " << mail() << endl << endl;
        stream << body() << endl;
        stream << "----------------------------------------" << endl;

        file.close();
    }
}


/* get result code from 2ch tag or title. */  /* private */
int WriteView::resultCode(const QString& response) const
{
    // see also libkita/kita-utf8.h
    QString errstr = QString::fromUtf8(KITAUTF8_WRITEERROR);
    QString truestr = QString::fromUtf8(KITAUTF8_WRITETRUE);
    QString ckstr = QString::fromUtf8(KITAUTF8_WRITECOOKIE);

    QRegExp regexp("<!-- 2ch_X:(.*) -->");
    int pos = regexp.indexIn(response);

    if (pos != -1) {
        // get code from 2ch tag
        QString k2ch_X = regexp.cap(1);

        if (k2ch_X == "true") return K2ch_True;
        if (k2ch_X == "false") return K2ch_False;
        if (k2ch_X == "error") return K2ch_Error;
        if (k2ch_X == "check") return K2ch_Check;
        if (k2ch_X == "cookie") return K2ch_Cookie;

        return K2ch_Unknown;
    } else {
        // get code from title
        QString title = resultTitle(response);
        if (title.isEmpty()) return K2ch_Unknown;

        if (title.contains(errstr)) return K2ch_Error;
        if (title.contains(truestr)) return K2ch_True;
        if (title.contains(ckstr)) return K2ch_Cookie;

        // for Flash CGI
        if (m_bbstype == Board_FlashCGI) {
            if (title.contains("ERROR!!")) {
                return K2ch_Error;
            } else {
                return K2ch_True;
            }
        }

        // for JBBS.  adhoc...
        if (m_bbstype == Board_JBBS) {

            // x-euc-jp & euc-jp
            if (response.contains("euc-jp")) {
                return K2ch_True;
            } else {
                return K2ch_Error;
            }
        }

        return K2ch_Unknown;
    }
}


/* private */
QString WriteView::resultMessage(const QString& response) const
{
    QRegExp tags("(<[^<]+>|</[^<]+>)");

    if (m_bbstype == Board_FlashCGI) {

        QRegExp regexp("<blockquote>(.*)</blockquote>");
        int pos = regexp.indexIn(response);
        if (pos != -1) {
            return regexp.cap(1).replace("<br>", "\n");
        }

        return QString();
    }

    {
        QRegExp bold_start("<b>");
        QRegExp bold_end("</b>");
        int startPos = bold_start.indexIn(response);
        if (startPos != -1) {
            startPos += 3;
            int endPos = bold_end.indexIn(response);
            if (endPos != -1 && startPos < endPos) {
                QString message = response.mid(startPos, endPos - startPos);
                message.remove(tags);
                return message;
            }
        }
    }

    QRegExp body_regexp("<body>([^<]*)</body>");
    if (body_regexp.indexIn(response) != -1) {
        return body_regexp.cap(1);
    }

    return QString();
}


/* private */
QString WriteView::resultTitle(const QString& response) const
{
    QRegExp regexp("<title>(.*)</title>");
    int pos = regexp.indexIn(response, Qt::CaseInsensitive);
    if (pos != -1) {
        return regexp.cap(1);
    } else {
        return QString();
    }
}
